# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import logging
from collections.abc import Sequence
from hashlib import sha256
from typing import Any
from typing import Optional
from typing import cast

import datarobot as dr
from airflow.exceptions import AirflowException
from airflow.utils.context import Context
from datarobot.enums import FileLocationType
from datarobot.models import Recipe
from datarobot.utils.source import parse_source_type

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator
from datarobot_provider.operators.base_datarobot_operator import BaseUseCaseEntityOperator

# Time in seconds after which dataset uploading is considered unsuccessful.
DATAROBOT_MAX_WAIT_SEC = 3600


class GetDataStoreOperator(BaseDatarobotOperator):
    """Get a DataRobot data store id by data connection name.
    You have to create a DataRobot data connection in advance
    at https://app.datarobot.com/account/data-connections page.

    :param data_connection: unique, case-sensitive data connection name as you can see it at DataRobot.
    :type data_connection: str
    :return: Data store ID.
    :rtype: str
    """

    template_fields: Sequence[str] = ["data_connection"]

    def __init__(
        self,
        *,
        data_connection: str = "{{ params.data_connection }}",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.data_connection = data_connection

    def execute(self, context: Context) -> Any:
        for datastore in dr.DataStore.list(name=self.data_connection):
            if datastore.canonical_name == self.data_connection:
                break

        else:
            raise AirflowException(f"Connection {self.data_connection} was not found.")

        return datastore.id


class UploadDatasetOperator(BaseUseCaseEntityOperator):
    """
    Uploading local file to DataRobot AI Catalog and return Dataset ID.
    :param file_path: The path to the file.
    :type file_path: str, optional
    :param file_path_param: DEPRECATED. Name of the parameter in the configuration to use as file_path, defaults to `dataset_file_path`
    :type file_path_param: str, optional
    :param use_case_id: ID of the use case to add the dataset into.
    :type use_case_id: str, optional
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: DataRobot AI Catalog dataset ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["file_path", "file_path_param", "use_case_id"]

    def __init__(
        self,
        *,
        file_path: str = "{{ params.dataset_file_path | default('') }}",  # Don't use any *default* after *dataset_file_path* is finally removed.
        file_path_param: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.file_path = file_path
        self.file_path_param = file_path_param

        if self.file_path_param is not None:
            self._file_path_param_is_deprecated()

    def execute(self, context: Context) -> str:
        # Upload Dataset to AI Catalog
        self.log.info("Upload Dataset to AI Catalog")
        if not self.file_path and self.file_path_param:
            self._file_path_param_is_deprecated()
            self.file_path = context["params"][self.file_path_param]

        source_type = parse_source_type(self.file_path)
        if source_type == FileLocationType.URL:
            ai_catalog_dataset: dr.Dataset = dr.Dataset.create_from_url(
                url=self.file_path, max_wait=DATAROBOT_MAX_WAIT_SEC
            )

        elif source_type == FileLocationType.PATH:
            ai_catalog_dataset = dr.Dataset.create_from_file(
                file_path=self.file_path, max_wait=DATAROBOT_MAX_WAIT_SEC
            )

        else:
            raise AirflowException(f"Unexpected file_path type: {source_type}")

        self.log.info(f"Dataset created: dataset_id={ai_catalog_dataset.id}")
        self.add_into_use_case(ai_catalog_dataset, context=context)

        return ai_catalog_dataset.id

    def _file_path_param_is_deprecated(self) -> None:
        self.log.warning(
            "**file_path_param** is deprecated. "
            f"Use `file_path={{{{ params.{self.file_path_param} }}}}` instead."
        )


class UpdateDatasetFromFileOperator(BaseDatarobotOperator):
    """
    Operator that creates a new Dataset version from a file.
    Returns when the new dataset version has been successfully uploaded.

    :param dataset_id: DataRobot AI Catalog dataset ID
    :type dataset_id: str, optional
    :param dataset_id_param: Name of the parameter in the configuration to use as dataset_id, defaults to `training_dataset_id`
    :type dataset_id_param: str, optional
    :param file_path: The path to the file to upload
    :type file_path: str, optional
    :param file_path_param: Name of the parameter in the configuration to use as file_path, defaults to `dataset_file_path`
    :type file_path_param: str, optional
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: DataRobot AI Catalog dataset version ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "dataset_id",
        "dataset_id_param",
        "file_path",
        "file_path_param",
    ]

    def __init__(
        self,
        *,
        dataset_id: Optional[str] = None,
        dataset_id_param: str = "training_dataset_id",
        file_path: Optional[str] = None,
        file_path_param: str = "dataset_file_path",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.dataset_id = dataset_id
        self.dataset_id_param = dataset_id_param
        self.file_path = file_path
        self.file_path_param = file_path_param

    def execute(self, context: Context) -> str:
        # If dataset_id not provided in constructor, then using dataset_id_param from json config
        dataset_id = (
            self.dataset_id
            if self.dataset_id is not None
            else context["params"][self.dataset_id_param]
        )

        # The path to the file.
        file_path = (
            self.file_path
            if self.file_path is not None
            else context["params"][self.file_path_param]
        )

        self.log.info(f"Update Dataset {dataset_id} in AI Catalog from the local file: {file_path}")
        ai_catalog_dataset = dr.Dataset.create_version_from_file(
            dataset_id=dataset_id,
            file_path=file_path,
            max_wait=DATAROBOT_MAX_WAIT_SEC,
        )
        self.log.info(
            f"Dataset updated: dataset_id={ai_catalog_dataset.id}, dataset_version_id:{ai_catalog_dataset.version_id}"
        )

        return ai_catalog_dataset.version_id


class CreateDatasetFromDataStoreOperator(BaseDatarobotOperator):
    """
    Loading dataset from JDBC Connection to DataRobot AI Catalog and return Dataset ID.

    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: DataRobot AI Catalog dataset ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["data_store_id", "credential_id"]

    def __init__(self, *, data_store_id: str, credential_id: Optional[str] = None, **kwargs: Any):
        super().__init__(**kwargs)
        self.data_store_id = data_store_id
        self.credential_id = credential_id

    def execute(self, context: Context) -> str:
        dataset_name = context["params"]["dataset_name"]

        data_source = None

        for dr_source_item in dr.DataSource.list():
            if dr_source_item.canonical_name == dataset_name:
                data_source = dr_source_item
                self.log.info(f"Found existing DataSource:{dataset_name}, id={data_source.id}")
                break

        # Creating DataSourceParameters:
        if "query" in context["params"] and context["params"]["query"]:
            # using sql statement if provided:
            params = dr.DataSourceParameters(query=context["params"]["query"])
        else:
            # otherwise using schema and table:
            params = dr.DataSourceParameters(
                schema=context["params"]["table_schema"], table=context["params"]["table_name"]
            )

        if data_source is None:
            # Adding data_store_id to params (required for DataSource creation):
            params.data_store_id = self.data_store_id
            # Creating DataSource using params with data_store_id
            self.log.info(f"Creating DataSource: {dataset_name}")
            data_source = dr.DataSource.create(
                data_source_type="jdbc", canonical_name=dataset_name, params=params
            )
            self.log.info(f"DataSource:{dataset_name} successfully created, id={data_source.id}")

        # Checking if there are any changes in params:
        elif params != data_source.params:
            # If params in changed, updating data source:
            self.log.info(f"Updating DataSource:{dataset_name} with new params")
            data_source.update(canonical_name=dataset_name, params=params)
            self.log.info(f"DataSource:{dataset_name} successfully updated, id={data_source.id}")

        self.log.info(f"Creating Dataset from DataSource: {dataset_name}")
        ai_catalog_dataset: dr.Dataset = dr.Dataset.create_from_data_source(
            data_source_id=data_source.id,
            credential_id=self.credential_id,
            persist_data_after_ingestion=context["params"]["persist_data_after_ingestion"],
            do_snapshot=context["params"]["do_snapshot"],
            max_wait=DATAROBOT_MAX_WAIT_SEC,
        )
        self.log.info(f"Dataset created: dataset_id={ai_catalog_dataset.id}")
        return ai_catalog_dataset.id


class CreateDatasetFromRecipeOperator(BaseUseCaseEntityOperator):
    """Create a dataset based on a wrangling recipe.
    The dataset can be dynamic or a snapshot depending on the mandatory *do_snapshot* parameter.
    The dataset is added into the Use Case if use_case_id is specified.

    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :param recipe_id: Wrangling or Feature Discovery Recipe Id
    :type recipe_id: str
    :param do_snapshot: *True* to download and store whole dataframe into DataRobot AI Catalog. *False* to create a dynamic dataset.
    :type do_snapshot: bool
    :param dataset_name: Name of the new dataset.
    :type dataset_name: str
    :param materialization_catalog: Data store catalog (database) to upload the wrangled data into.
    :type materialization_catalog: str
    :param materialization_schema: The database schema to upload the wrangled data into.
    :type materialization_schema: str
    :param materialization_table: The database table to upload the wrangled data into.
    :type materialization_table: str
    :param use_case_id: ID of the use case to add the dataset into.
    :type use_case_id: str or None
    :return: DataRobot AI Catalog dataset ID
    :rtype: str
    """

    template_fields = [
        "recipe_id",
        "use_case_id",
        "dataset_name",
        "materialization_catalog",
        "materialization_schema",
        "materialization_table",
    ]

    def __init__(
        self,
        *,
        recipe_id: str,
        do_snapshot: bool,
        dataset_name: Optional[str] = "{{ params.dataset_name | default('') }}",
        materialization_catalog: Optional[
            str
        ] = "{{ params.materialization_catalog | default('') }}",
        materialization_schema: Optional[str] = "{{ params.materialization_schema | default('') }}",
        materialization_table: Optional[str] = "{{ params.materialization_table | default('') }}",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.recipe_id = recipe_id
        self.do_snapshot = do_snapshot
        self.dataset_name = dataset_name
        self.materialization_catalog = materialization_catalog
        self.materialization_schema = materialization_schema
        self.materialization_table = materialization_table

    def _get_materialization_destination(
        self,
    ) -> Optional[dr.models.dataset.MaterializationDestination]:
        if self.materialization_table:
            return dr.models.dataset.MaterializationDestination(
                catalog=self.materialization_catalog or None,  # type: ignore[typeddict-item]
                schema=self.materialization_schema or None,  # type: ignore[typeddict-item]
                table=self.materialization_table,
            )

        return None

    def _get_dataset_name(
        self,
        materialization_destination: Optional[dr.models.dataset.MaterializationDestination],
    ) -> Optional[str]:
        return self.dataset_name or (
            materialization_destination and materialization_destination["table"]
        )

    def execute(self, context: Context) -> str:
        recipe = Recipe.get(self.recipe_id)
        if recipe.dialect == dr.enums.DataWranglingDialect.SPARK and not self.do_snapshot:
            raise AirflowException(
                "Dynamic datasets are not suitable for 'spark' recipes. "
                "Please, either specify do_snapshot=True for the operator or use another recipe."
            )

        materialization_destination = self._get_materialization_destination()
        dataset_name = self._get_dataset_name(materialization_destination)

        dataset: dr.Dataset = dr.Dataset.create_from_recipe(
            recipe,
            name=dataset_name,
            do_snapshot=self.do_snapshot,
            persist_data_after_ingestion=True,
            materialization_destination=materialization_destination,
        )

        logging.info(
            '%s dataset "%s" created.',
            "Snapshot" if self.do_snapshot else "Dynamic",
            dataset.name,
        )
        self.add_into_use_case(dataset, context=context)

        return dataset.id


class CreateDatasetVersionOperator(BaseDatarobotOperator):
    """
    Creating new version of existing dataset in AI Catalog and return dataset version ID.

    :param dataset_id: DataRobot AI Catalog dataset ID
    :type dataset_id: str
    :param datasource_id: existing DataRobot datasource ID
    :type datasource_id: str
    :param credential_id: existing DataRobot credential ID
    :type credential_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: DataRobot AI Catalog dataset version ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["dataset_id", "datasource_id", "credential_id"]

    def __init__(
        self,
        *,
        dataset_id: str,
        datasource_id: str,
        credential_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.dataset_id = dataset_id
        self.datasource_id = datasource_id
        self.credential_id = credential_id

    def execute(self, context: Context) -> str:
        self.log.debug(
            f"Creation new version of dataset: dataset_id={self.dataset_id}, "
            f"using datasource: datasource_id={self.datasource_id}, "
            f"with credentials: credentials_id={self.credential_id}."
        )

        ai_catalog_dataset = dr.Dataset.create_version_from_data_source(
            dataset_id=self.dataset_id,
            data_source_id=self.datasource_id,
            credential_id=self.credential_id,
            max_wait=DATAROBOT_MAX_WAIT_SEC,
        )

        self.log.info(
            f"Dataset version created: dataset_id={ai_catalog_dataset.id},"
            f" version_id={ai_catalog_dataset.version_id}"
        )

        return ai_catalog_dataset.version_id


class CreateDatasetFromProjectOperator(BaseDatarobotOperator):
    """
    Create a new AI Catalog Dataset from existing project data.
    :param project_id: DataRobot project ID
    :type project_id: str
    :param datasource_id: existing DataRobot datasource ID
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: DataRobot AI Catalog dataset ID
    :rtype: str
    """

    template_fields: Sequence[str] = ["project_id"]

    def __init__(
        self,
        *,
        project_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id

    def execute(self, context: Context) -> str:
        dataset: dr.Dataset = dr.Dataset.create_from_project(project_id=self.project_id)
        self.log.info(f"Dataset created: dataset_id={dataset.id}")
        return dataset.id


class CreateOrUpdateDataSourceOperator(BaseDatarobotOperator):
    """
    Get an existing data source by name and update it if any of *table_schema*, *table_name*, *query* are specified.
    Create a new data source if there is no existing one with the specified name.

    :param data_store_id: DataRobot data store ID
    :type data_store_id: str
    :param dataset_name: Data source canonical name to create or update.
    :type dataset_name: Optional[str]
    :param table_schema: Database schema name.
    :type table_schema: Optional[str]
    :param table_name: Database table name.
    :type table_name: Optional[str]
    :param query: Database table name.
    :type query: Optional[str]

    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: DataRobot AI Catalog data source ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "data_store_id",
        "dataset_name",
        "table_name",
        "table_schema",
        "query",
    ]

    def __init__(
        self,
        *,
        data_store_id: str,
        dataset_name: Optional[str] = "{{ params.get('dataset_name', '') }}",
        table_schema: Optional[str] = "{{ params.get('table_schema', '') }}",
        table_name: Optional[str] = "{{ params.get('table_name', '') }}",
        query: Optional[str] = "{{ params.get('query', '') }}",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.data_store_id = data_store_id
        self.dataset_name = dataset_name
        self.table_schema = table_schema
        self.table_name = table_name
        self.query = query

    def execute(self, context: Context) -> Optional[str]:
        self.log.debug(f"Trying to get existing DataStore by data_store_id={self.data_store_id}")
        data_store = dr.DataStore.get(data_store_id=self.data_store_id)
        self.log.debug(f"Found existing DataStore: {data_store.canonical_name}, id={data_store.id}")

        if not self.dataset_name:
            self.dataset_name = self._get_default_data_source_name(cast(str, data_store.id))
            self.log.info("Use default name for the data source: %s", self.dataset_name)

        # Creating DataSourceParameters:
        if self.query:
            # using sql statement if provided:
            params = dr.DataSourceParameters(query=self.query)
        elif self.table_name:
            # otherwise using schema and table:
            params = dr.DataSourceParameters(schema=self.table_schema, table=self.table_name)

        else:
            # or only search by name.
            params = None

        self.log.debug(f"Trying to get existing DataSource by name={self.dataset_name}")
        for data_source in dr.DataSource.list():
            if data_source.canonical_name == self.dataset_name:
                self.log.info(f"Found existing DataSource:{self.dataset_name}, id={data_source.id}")
                if params is not None and params != data_source.params:
                    # If params in changed, updating data source:
                    self.log.info(f"Updating DataSource:{self.dataset_name} with new params")
                    data_source.update(canonical_name=self.dataset_name, params=params)
                    self.log.info(
                        f"DataSource:{self.dataset_name} successfully updated, id={data_source.id}"
                    )
                break
        else:
            if params is None:
                raise AirflowException(
                    f"{self.dataset_name} data source was not found. "
                    "Set *table_schema* and *table_name* or a *query* parameter "
                    "to create a new one instead."
                )

            # Adding data_store_id to params (required for DataSource creation):
            params.data_store_id = data_store.id
            # Creating DataSource using params with data_store_id
            self.log.info(f"Creating DataSource: {self.dataset_name}")
            data_source = dr.DataSource.create(
                data_source_type="jdbc", canonical_name=self.dataset_name, params=params
            )
            self.log.info(
                f"DataSource:{self.dataset_name} successfully created, id={data_source.id}"
            )

        return data_source.id

    def _get_default_data_source_name(self, data_store_id: str) -> str:
        """Build default name based on the data source params."""
        parts = ["Airflow", data_store_id]

        if self.query:
            parts += ["q", sha256(self.query.encode()).hexdigest()]

        else:
            parts_additions = ["t"]
            if self.table_schema:
                parts_additions.append(self.table_schema)
            if self.table_name:
                parts_additions.append(self.table_name)
            parts += parts_additions

        return "-".join(parts)



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections.abc import Sequence
from typing import TYPE_CHECKING
from typing import Any

import datarobot as dr
from airflow.utils.context import Context

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator

if TYPE_CHECKING:
    from datarobot.models.deployment.deployment import SegmentAnalysisSettings


class GetSegmentAnalysisSettingsOperator(BaseDatarobotOperator):
    """
    Get segment analysis settings for a deployment.

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["deployment_id"]

    def __init__(
        self,
        *,
        deployment_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id

    def execute(self, context: Context) -> "SegmentAnalysisSettings":
        self.log.info(f"Get Deployment for deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(deployment_id=self.deployment_id)

        segment_analysis_settings = deployment.get_segment_analysis_settings()

        return segment_analysis_settings


class UpdateSegmentAnalysisSettingsOperator(BaseDatarobotOperator):
    """
    Updates segment analysis settings for a deployment.

    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["deployment_id"]

    def __init__(
        self,
        *,
        deployment_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id

    def execute(self, context: Context) -> None:
        self.log.info(f"Getting Deployment for deployment_id={self.deployment_id}")
        deployment = dr.Deployment.get(deployment_id=self.deployment_id)

        current_segment_analysis_settings = deployment.get_segment_analysis_settings()

        segment_analysis_enabled = context["params"].get(
            "segment_analysis_enabled", current_segment_analysis_settings["enabled"]
        )
        segment_analysis_attributes = context["params"].get(
            "segment_analysis_attributes", current_segment_analysis_settings["attributes"]
        )

        if (segment_analysis_enabled != current_segment_analysis_settings["enabled"]) or (
            segment_analysis_attributes != current_segment_analysis_settings["attributes"]
        ):
            self.log.debug(
                f"Trying to update segment analysis settings for deployment_id={self.deployment_id}"
            )
            deployment.update_segment_analysis_settings(
                segment_analysis_enabled=segment_analysis_enabled,
                segment_analysis_attributes=segment_analysis_attributes,
            )
            self.log.info(
                f"Deployment segment analysis settings updated for deployment_id={self.deployment_id}"
            )
        else:
            self.log.info(
                f"No need to update segment analysis settings for deployment_id={self.deployment_id}"
            )



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

import datarobot as dr
from airflow.decorators import dag

from datarobot_provider.operators.data_prep import CreateWranglingRecipeOperator
from datarobot_provider.operators.data_registry import CreateDatasetFromRecipeOperator
from datarobot_provider.operators.data_registry import UploadDatasetOperator
from datarobot_provider.operators.datarobot import CreateProjectOperator
from datarobot_provider.operators.datarobot import GetOrCreateUseCaseOperator
from datarobot_provider.operators.datarobot import SelectBestModelOperator
from datarobot_provider.operators.datarobot import TrainModelsOperator
from datarobot_provider.operators.model_registry import CreateRegisteredModelVersionOperator
from datarobot_provider.sensors.datarobot import AutopilotCompleteSensor

"""
Example of Aiflow DAG for DataRobot data preparation and model training.
Configurable parameters for this dag:
* dataset_file_path - URL or a local path for a csv/parquet/xlsx file.
* project_name - name of the project as displayed in DataRobot UI.
* autopilot_settings - a dictionary with the modelling project autopilot settings.
"""


@dag(
    schedule=None,
    tags=["example", "csv", "wrangling", "modeling"],
    params={
        "dataset_file_path": "https://s3.amazonaws.com/datarobot_public_datasets/10k_diabetes.csv",
        "project_name": "hospital-readmissions-example",
        "autopilot_settings": {"target": "readmitted", "mode": "quick", "max_wait": 3600},
    },
)
def dataprep_wrangler_autopilot():
    # Create a Use Case to keep all subsequent assets. Default name is "Airflow"
    create_use_case = GetOrCreateUseCaseOperator(task_id="create_use_case", set_default=True)

    # Upload the data into Data Registry.
    upload_dataset = UploadDatasetOperator(task_id="upload_dataset")

    # Define data preparation.
    create_recipe = CreateWranglingRecipeOperator(
        task_id="create_recipe",
        dataset_id=upload_dataset.output,
        dialect=dr.enums.DataWranglingDialect.SPARK,
        # See the list of available *operation* options in the DtaRobot API documentation:
        # https://docs.datarobot.com/en/docs/api/reference/public-api/data_wrangling.html#schemaoneofdirective
        # General *operation* structure is:
        # {"directive": <One of dr.enums.WranglingOperations>, "arguments": <dictionary>}
        operations=[
            {
                "directive": "rename-columns",
                "arguments": {
                    "columnMappings": [
                        {"originalName": "admission_type_id", "newName": "AdmissionType"},
                        {
                            "originalName": "discharge_disposition_id",
                            "newName": "DischargeDisposition",
                        },
                        {"originalName": "admission_source_id", "newName": "AdmissionSource"},
                    ]
                },
            },
            {
                "directive": "replace",
                "arguments": {
                    "origin": "AdmissionType",
                    "searchFor": "",
                    "replacement": "Not Available",
                    "matchMode": "exact",
                },
            },
            {
                "directive": "replace",
                "arguments": {
                    "searchFor": r"\\[(\\d+).*",
                    "replacement": r"$1",
                    "origin": "age",
                    "matchMode": "regex",
                },
            },
            {
                "directive": "compute-new",
                "arguments": {
                    "expression": "CAST(`age` AS Integer)",
                    "newFeatureName": "int-age",
                },
            },
            {
                "directive": "drop-columns",
                "arguments": {
                    "columns": [
                        "citoglipton",
                        "acetohexamide",
                        "miglitol",
                        "troglitazone",
                        "examide",
                        "age",
                    ]
                },
            },
        ],
    )

    # Apply data preparation and save the modified data in the Data Registry.
    publish_recipe = CreateDatasetFromRecipeOperator(
        task_id="publish_recipe",
        recipe_id=create_recipe.output,
        do_snapshot=True,
    )

    # Create a new Project.
    create_project = CreateProjectOperator(
        task_id="create_project",
        dataset_id=publish_recipe.output,
    )

    # Launch modeling autopilot.
    train_models = TrainModelsOperator(task_id="train_models", project_id=create_project.output)

    # Wait for the autopilot completion.
    autopilot_complete_sensor = AutopilotCompleteSensor(
        task_id="check_autopilot_complete",
        project_id=create_project.output,
    )

    # select best model
    select_best_model = SelectBestModelOperator(
        task_id="select_best_model",
        project_id=create_project.output,
        metric="RMSE",
    )
    # register model
    register_model = CreateRegisteredModelVersionOperator(
        task_id="register_model",
        model_version_params={
            "model_type": "leaderboard",
            "model_id": select_best_model.output,
            "name": "Highest readmitted score test",
            "registered_model_name": "Highest readmitted score test {{ ts }}",
        },
    )

    (
        create_use_case
        >> upload_dataset
        >> create_recipe
        >> publish_recipe
        >> create_project
        >> train_models
        >> autopilot_complete_sensor
        >> select_best_model
        >> register_model
    )


dataprep_wrangler_autopilot()



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime

from airflow.decorators import dag
from airflow.decorators import task
from datarobot import PredictJob

from datarobot_provider.hooks.datarobot import DataRobotHook
from datarobot_provider.operators.model_predictions import AddExternalDatasetOperator
from datarobot_provider.operators.model_predictions import RequestModelPredictionsOperator
from datarobot_provider.sensors.model_insights import DataRobotJobSensor


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example", "dataset"],
)
def compute_model_predictions(
    project_id=None,
    model_id=None,
    dataset_id=None,
):
    if not project_id:
        raise ValueError("Invalid or missing `project_id` value")
    if not model_id:
        raise ValueError("Invalid or missing `model_id` value")
    if not dataset_id:
        raise ValueError("Invalid or missing `dataset_id` value")

    add_external_dataset_op = AddExternalDatasetOperator(
        task_id="add_external_dataset",
        project_id=project_id,
        dataset_id=dataset_id,
    )

    request_model_predictions_op = RequestModelPredictionsOperator(
        task_id="request_model_predictions",
        project_id=project_id,
        model_id=model_id,
        external_dataset_id=add_external_dataset_op.output,
    )

    model_predictions_sensor = DataRobotJobSensor(
        task_id="model_predictions_complete",
        project_id=project_id,
        job_id=request_model_predictions_op.output,
        poke_interval=5,
        timeout=3600,
    )

    @task(task_id="example_custom_python_code")
    def using_custom_python_code(project_id, predict_job_id, datarobot_conn_id="datarobot_default"):
        """Example of using custom python code:"""

        # Initialize DataRobot client
        DataRobotHook(datarobot_conn_id).run()

        # Should be used only with a very small datasets (not higher than a few megabytes).
        output_df = PredictJob.get_predictions(project_id=project_id, predict_job_id=predict_job_id)

        # Put your logic with the model predictions output dataframe here, for example:
        return output_df["positive_probability"].mean() > 0.5

    example_custom_python_code = using_custom_python_code(
        project_id=project_id, predict_job_id=request_model_predictions_op.output
    )

    (
        add_external_dataset_op
        >> request_model_predictions_op
        >> model_predictions_sensor
        >> example_custom_python_code
    )


compute_model_predictions_dag = compute_model_predictions()

if __name__ == "__main__":
    compute_model_predictions_dag.test()



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

import datarobot as dr
from airflow.decorators import dag

from datarobot_provider.operators.data_prep import CreateWranglingRecipeOperator
from datarobot_provider.operators.data_registry import CreateDatasetFromRecipeOperator
from datarobot_provider.operators.data_registry import CreateOrUpdateDataSourceOperator
from datarobot_provider.operators.data_registry import GetDataStoreOperator
from datarobot_provider.operators.datarobot import CreateProjectOperator
from datarobot_provider.operators.datarobot import GetOrCreateUseCaseOperator
from datarobot_provider.operators.datarobot import GetProjectBlueprintsOperator
from datarobot_provider.operators.datarobot import TrainModelsOperator
from datarobot_provider.operators.model_insights import ComputeShapPreviewOperator
from datarobot_provider.operators.model_registry import CreateRegisteredModelVersionOperator
from datarobot_provider.operators.model_training import TrainModelOperator
from datarobot_provider.sensors.model_training import ModelTrainingJobSensor

"""
Example of Aiflow DAG to apply db data transformations and train an xgboost model.
Configurable parameters for this dag:
* data_connection - database connection name you can find at https://app.datarobot.com/account/data-connections
* table_schema - the database schema to use
* primary_table - the table to start with
* secondary_table - the table to JOIN to the original table
* project_name - name of the project as displayed in DataRobot UI.
* autopilot_settings - a dictionary with the modelling project autopilot settings.
"""


@dag(
    schedule=None,
    tags=["example", "snowflake", "wrangling", "modeling"],
    params={
        "data_connection": "Demo Connection",
        "table_schema": "TRIAL_READONLY",
        "primary_table": "LENDING_CLUB_TRAINING",
        "secondary_table": "LENDING_CLUB_TRANSACTIONS",
        "project_name": "Lending Club",
        "autopilot_settings": {"target": "BadLoan", "mode": "manual", "max_wait": 3600},
    },
)
def model_training_xgboost():
    # Create a Use Case to keep all subsequent assets. Default name is "Airflow"
    create_use_case = GetOrCreateUseCaseOperator(task_id="create_use_case", set_default=True)

    # Define input data
    get_data_store = GetDataStoreOperator(task_id="get_data_store")
    define_transactions_table = CreateOrUpdateDataSourceOperator(
        task_id="define_transactions_table",
        data_store_id=get_data_store.output,
        table_name="{{ params.secondary_table }}",
    )

    # Define data preparation:
    # * Join `LENDING_CLUB_TRANSACTIONS` table,
    # * cast the `Amount` column to decimal
    # * and apply different types of aggregation.
    create_recipe = CreateWranglingRecipeOperator(
        task_id="create_recipe",
        data_store_id=get_data_store.output,
        table_name="{{ params.primary_table }}",
        dialect=dr.enums.DataWranglingDialect.SNOWFLAKE,
        operations=[
            {
                "directive": "join",
                "arguments": {
                    "leftKeys": ["CustomerID"],
                    "rightKeys": ["CustomerID"],
                    "joinType": "left",
                    "source": "table",
                    "rightDataSourceId": define_transactions_table.output,
                },
            },
            {
                "directive": "replace",
                "arguments": {
                    "origin": "Amount",
                    "searchFor": "[$,]",
                    "replacement": "",
                    "matchMode": "regex",
                    "isCaseSensitive": False,
                },
            },
            {
                "directive": "compute-new",
                "arguments": {
                    "expression": 'CAST("Amount" AS Decimal(10, 2))',
                    "newFeatureName": "decimal amount",
                },
            },
            {
                "directive": "aggregate",
                "arguments": {
                    "groupBy": ["CustomerID", "BadLoan", "date"],
                    "aggregations": [
                        {
                            "feature": "decimal amount",
                            "functions": ["sum", "avg", "stddev", "min", "max", "median"],
                        }
                    ],
                },
            },
        ],
    )

    # Apply data preparation and save the modified data in the Data Registry.
    publish_recipe = CreateDatasetFromRecipeOperator(
        task_id="publish_recipe",
        recipe_id=str(create_recipe.output),
        do_snapshot=True,
    )

    # Create a new Project.
    create_project = CreateProjectOperator(
        task_id="create_project",
        dataset_id=str(publish_recipe.output),
    )
    project_id = str(create_project.output)

    # Launch modeling autopilot in manual mode.
    start_modeling = TrainModelsOperator(task_id="train_models", project_id=project_id)

    # Get the blueprint id of an xgboost model.
    get_blueprint_id = GetProjectBlueprintsOperator(
        task_id="get_blueprint_id",
        project_id=project_id,
        filter_model_type="extreme gradient boosted",
    )

    trained_model = TrainModelOperator(
        task_id="train_model",
        project_id=project_id,
        blueprint_id=str(get_blueprint_id.output),
    )

    trained_model_sensor = ModelTrainingJobSensor(
        task_id="model_training_complete",
        project_id=project_id,
        job_id=str(trained_model.output),
        poke_interval=5,
        timeout=3600,
    )

    insights = ComputeShapPreviewOperator(
        task_id="compute_shap_insights",
        model_id=str(trained_model_sensor.output),
    )

    # register model
    registered_model_name = f"Highest readmitted score {str(trained_model_sensor.output)}"
    register_model = CreateRegisteredModelVersionOperator(
        task_id="register_model",
        model_version_params={
            "model_type": "leaderboard",
            "model_id": str(trained_model_sensor.output),
            "name": registered_model_name,
            "registered_model_name": registered_model_name,
        },
    )

    (
        create_use_case
        >> get_data_store
        >> define_transactions_table
        >> create_recipe
        >> publish_recipe
        >> create_project
        >> start_modeling
        >> get_blueprint_id
        >> trained_model
        >> trained_model_sensor
        >> insights
        >> register_model
    )


model_training_xgboost()



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

import datetime
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

import datarobot as dr
from airflow.exceptions import AirflowException
from airflow.utils.context import Context
from datarobot import DataStore
from datarobot.enums import DataWranglingDataSourceTypes
from datarobot.enums import DataWranglingDialect
from datarobot.enums import DownsamplingOperations
from datarobot.enums import RecipeInputType
from datarobot.models import DataSourceInput
from datarobot.models import JDBCTableDataSourceInput
from datarobot.models import Recipe
from datarobot.models import RecipeDatasetInput
from datarobot.models.dataset import Dataset
from datarobot.models.recipe_operation import DownsamplingOperation
from datarobot.models.recipe_operation import RandomSamplingOperation
from datarobot.models.recipe_operation import WranglingOperation

from datarobot_provider.operators.base_datarobot_operator import BaseUseCaseEntityOperator


class CreateWranglingRecipeOperator(BaseUseCaseEntityOperator):
    """Create a Wrangling Recipe

    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :param use_case_id: Use Case ID to create the recipe in.
    :param dataset_id: The dataset to wrangle
    :param dialect: SQL dialect to apply while wrangling.
    :param recipe_name: New recipe name.
    :param recipe_description: New recipe description.
    :param operations: Wrangling operations to apply.
    :param downsampling_directive: Downsampling method to apply. *None* for non downsampling.
    :param downsampling_arguments_param: Downsampling arguments.

    """

    template_fields: Sequence[str] = [
        "use_case_id",
        "dataset_id",
        "data_store_id",
        "table_schema",
        "table_name",
        "dialect",
        "recipe_name",
        "recipe_description",
        "operations",
        "downsampling_directive",
        "downsampling_arguments",
    ]
    template_fields_renderers: dict[str, str] = {
        "use_case_id": "string",
        "dataset_id": "string",
        "data_store_id": "string",
        "table_schema": "string",
        "table_name": "string",
        "dialect": "string",
        "recipe_name": "string",
        "recipe_description": "string",
        "operations": "json",
        "downsampling_directive": "string",
        "downsampling_arguments": "json",
    }

    def __init__(
        self,
        *,
        dataset_id: Optional[str] = None,
        data_store_id: Optional[str] = None,
        table_schema: Optional[str] = "{{ params.get('table_schema', '') }}",
        table_name: Optional[str] = "{{ params.get('table_name', '') }}",
        dialect: DataWranglingDialect,
        recipe_name: Optional[str] = None,
        recipe_description: Optional[str] = "Created with Apache-Airflow",
        operations: Optional[List[dict]] = None,
        downsampling_directive: Optional[DownsamplingOperations] = None,
        downsampling_arguments: Optional[dict] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.dataset_id = dataset_id
        self.data_store_id = data_store_id
        self.table_schema = table_schema
        self.table_name = table_name
        self.dialect = dialect
        self.recipe_name = recipe_name
        self.recipe_description = recipe_description
        self.operations = operations or []
        self.downsampling_directive = downsampling_directive
        self.downsampling_arguments = downsampling_arguments

    def validate(self) -> None:
        if self.dataset_id and self.data_store_id:
            raise AirflowException(
                "You have to specify either dataset_id or data_store_id. Not both."
            )

    def execute(self, context: Context) -> str:
        use_case: dr.UseCase = self.get_use_case(context, required=True)  # type: ignore[assignment]

        if self.dataset_id:
            self.log.info("Working with dataset_id=%s", self.dataset_id)

            dataset = Dataset.get(self.dataset_id)
            recipe = Recipe.from_dataset(
                use_case, dataset, dialect=DataWranglingDialect(self.dialect)
            )

        elif self.data_store_id:
            if not self.table_name:
                raise AirflowException(
                    "*table_name* parameter must be specified "
                    "when working with a data store (db connection)."
                )

            self.log.info("Working with data_store_id=%s", self.data_store_id)
            data_store = DataStore.get(self.data_store_id)
            if not (data_store.type and data_store.type in iter(DataWranglingDataSourceTypes)):
                raise AirflowException(f"Unexpected data store type: {data_store.type}")

            data_source_canonical_name = self._generate_data_source_canonical_name()

            recipe = Recipe.from_data_store(
                use_case,
                data_store,
                data_source_type=DataWranglingDataSourceTypes(data_store.type),
                dialect=DataWranglingDialect(self.dialect),
                data_source_inputs=[
                    DataSourceInput(
                        canonical_name=data_source_canonical_name,
                        schema=self.table_schema,
                        table=self.table_name,
                        sampling=RandomSamplingOperation(1000, 0),
                    )
                ],
            )

        else:
            raise AirflowException("Please specify either dataset_id or data_store_id to wrangle.")

        self.log.info(
            '%s recipe id=%s created in use case "%s". Configuring...',
            self.dialect,
            recipe.id,
            use_case.name,
        )

        if self.operations:
            self._set_operations(recipe)
            self.log.info("%d operations set.", len(self.operations))

        if self.downsampling_directive is not None:
            client_downsampling = DownsamplingOperation(
                directive=DownsamplingOperations(self.downsampling_directive),
                arguments=self.downsampling_arguments,
            )
            Recipe.update_downsampling(recipe.id, client_downsampling)
            self.log.info("%s dowsnsampling set.", self.downsampling_directive)

        if self.recipe_name or self.recipe_description:
            data = {"description": self.recipe_description}
            if self.recipe_name:
                data["name"] = self.recipe_name

            dr.client.get_client().patch(f"recipes/{recipe.id}/", json=data)
            self.log.info("Recipe name/description set.")

        self.log.info("Recipe id=%s is ready.", recipe.id)
        return recipe.id

    def _generate_data_source_canonical_name(self) -> str:
        base_name = f"{self.table_name}-{datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')}"

        if self.table_schema:
            base_name = f"{self.table_schema}-{base_name}"

        return f"Airflow:{base_name}"

    def _set_operations(self, recipe: Recipe) -> None:
        secondary_inputs: Dict[str, Union[JDBCTableDataSourceInput, RecipeDatasetInput]] = {}

        if recipe.inputs[0].input_type == RecipeInputType.DATASOURCE:
            data_store_id = recipe.inputs[0].data_store_id  # type: ignore[union-attr]
            primary_dataset_id = None
        else:
            data_store_id = None
            primary_dataset_id = recipe.inputs[0].dataset_id  # type: ignore[union-attr]

        for operation_data in self.operations:
            if operation_data["directive"] == "join":
                if operation_data["arguments"].get("rightDataSourceId"):
                    secondary_inputs[operation_data["arguments"]["rightDataSourceId"]] = (
                        JDBCTableDataSourceInput(
                            input_type=RecipeInputType.DATASOURCE,
                            data_store_id=data_store_id,  # type: ignore[arg-type]
                            data_source_id=operation_data["arguments"]["rightDataSourceId"],
                        )
                    )

                else:
                    if not operation_data["arguments"].get("rightDatasetVersionId"):
                        dataset = Dataset.get(operation_data["arguments"]["rightDatasetId"])
                        operation_data["arguments"]["rightDatasetVersionId"] = dataset.version_id

                    if operation_data["arguments"]["rightDatasetId"] != primary_dataset_id:
                        secondary_inputs[operation_data["arguments"]["rightDatasetVersionId"]] = (
                            RecipeDatasetInput(
                                input_type=RecipeInputType.DATASET,
                                dataset_id=operation_data["arguments"]["rightDatasetId"],
                                dataset_version_id=operation_data["arguments"][
                                    "rightDatasetVersionId"
                                ],
                            )
                        )

        if secondary_inputs:
            inputs = recipe.inputs + list(secondary_inputs.values())
            Recipe.set_inputs(recipe.id, inputs)

        client_operations = [WranglingOperation.from_data(x) for x in self.operations]

        Recipe.set_operations(recipe.id, client_operations)


class CreateFeatureDiscoveryRecipeOperator(BaseUseCaseEntityOperator):
    """
    Create a Feature Discovery recipe.

    :param dataset_id: DataRobot Dataset Catalog ID to use as the primary dataset
    :type dataset_id: str
    :param use_case_id: ID of the use case to add the recipe to
    :type use_case_id: str
    :param dataset_definitions: list of dataset definitions
        Each element is a dict retrieved from DatasetDefinitionOperator operator
    :type dataset_definitions: Iterable[dict]
    :param relationships: list of relationships
        Each element is a dict retrieved from DatasetRelationshipOperator operator
    :type relationships: Iterable[dict]
    :param feature_discovery_settings: list of feature discovery settings, optional
        If not provided, it will be retrieved from DAG configuration params otherwise default
        settings will be used.
    :type feature_discovery_settings: Iterable[dict]
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: DataRobot Recipe ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "dataset_id",
        "use_case_id",
        "dataset_definitions",
        "relationships",
        "feature_discovery_settings",
    ]

    def __init__(
        self,
        *,
        dataset_id: str,
        dataset_definitions: Iterable[dict],
        relationships: Iterable[dict],
        feature_discovery_settings: Optional[Iterable[dict]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.dataset_id = dataset_id
        self.dataset_definitions = dataset_definitions
        self.relationships = relationships
        self.feature_discovery_settings = feature_discovery_settings

    def execute(self, context: Context) -> str:
        use_case: dr.UseCase = self.get_use_case(context, required=True)  # type: ignore[assignment]
        dataset = dr.Dataset.get(self.dataset_id)

        recipe = dr.models.Recipe.from_dataset(
            use_case,
            dataset,
            recipe_type=dr.enums.RecipeType.FEATURE_DISCOVERY,
        )

        # Get dataset version ID if it isn't defined by the user:
        for dataset_definition in self.dataset_definitions:
            if not dataset_definition.get("catalogVersionId"):
                secondary_dataset = dr.Dataset.get(dataset_definition["catalogId"])
                dataset_definition["catalogVersionId"] = secondary_dataset.version_id

        recipe_id = recipe.id
        recipe_config_id = recipe.settings.relationships_configuration_id  # type: ignore[union-attr]

        # Add secondary dataset configuration information into the Recipe config
        dr.RelationshipsConfiguration(recipe_config_id).replace(
            self.dataset_definitions,
            self.relationships,
            self.feature_discovery_settings,
        )

        self.log.info(f"Feature Discovery recipe created: recipe_id={recipe_id}")

        return recipe_id



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any
from typing import Tuple

from airflow import __version__ as AIRFLOW_VERSION
from airflow.exceptions import AirflowException
from airflow.hooks.base import BaseHook
from datarobot.client import Client
from datarobot.rest import RESTClientObject

from datarobot_provider import get_provider_info
from datarobot_provider.constants import DATAROBOT_CONN_ID


class DataRobotHook(BaseHook):
    """
    A hook that interacts with DataRobot via its public Python API library.

    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    """

    conn_name_attr = "datarobot_conn_id"
    default_conn_name = DATAROBOT_CONN_ID
    conn_type = "http"
    hook_name = "DataRobot"

    @staticmethod
    def get_connection_form_widgets() -> dict[str, Any]:
        """Returns connection widgets to add to connection form."""
        from flask_appbuilder.fieldwidgets import BS3PasswordFieldWidget
        from flask_appbuilder.fieldwidgets import BS3TextFieldWidget
        from flask_babel import lazy_gettext
        from wtforms import PasswordField
        from wtforms import StringField

        return {
            "extra__http__endpoint": StringField(
                lazy_gettext("DataRobot endpoint URL"),
                widget=BS3TextFieldWidget(),
                default="https://app.datarobot.com/api/v2",
            ),
            "extra__http__api_key": PasswordField(
                lazy_gettext("API Key"), widget=BS3PasswordFieldWidget()
            ),
        }

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        """Returns custom field behaviour."""
        return {
            "hidden_fields": ["host", "schema", "login", "password", "port", "extra"],
            "relabeling": {},
            "placeholders": {
                "extra__http__endpoint": "https://app.datarobot.com/api/v2",
                "extra__http__api_key": "your-api-key",
            },
        }

    def __init__(
        self,
        datarobot_conn_id: str = default_conn_name,
    ) -> None:
        super().__init__()
        self.datarobot_conn_id = datarobot_conn_id

    def get_conn(self) -> RESTClientObject:
        """Initializes a DataRobot client instance."""
        conn = self.get_connection(self.datarobot_conn_id)
        endpoint = conn.extra_dejson.get("extra__http__endpoint", "")
        if not endpoint:
            raise AirflowException("Endpoint is not defined")
        api_key = conn.extra_dejson.get("extra__http__api_key", "")
        if not api_key:
            raise AirflowException("API key is not defined")

        # Creating version-specific user agent suffix for collecting usage statistics and troubleshoot purposes:
        provider_package_name = get_provider_info().get("package-name")
        provider_versions = "".join(get_provider_info()["versions"])
        user_agent_suffix = "{}-{}-airflow-{}".format(
            provider_package_name, provider_versions, AIRFLOW_VERSION
        )
        self.log.info("Initialize DataRobot Client, user_agent_suffix:{}".format(user_agent_suffix))
        return Client(token=api_key, endpoint=endpoint, trace_context=user_agent_suffix)

    def run(self) -> Any:
        # Initialize DataRobot client
        return self.get_conn()

    def test_connection(self) -> Tuple[bool, str]:
        """Test HTTP Connection"""
        try:
            self.run()
            return True, "Connection successfully tested"
        except Exception as e:
            return False, str(e)



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections.abc import Sequence
from typing import Any
from typing import Optional

import datarobot as dr
from airflow.utils.context import Context

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator

DEFAULT_MAX_WAIT_SEC = 600


class AddExternalDatasetOperator(BaseDatarobotOperator):
    """
    Upload a new dataset from a catalog dataset to make predictions for a model
    :param project_id: DataRobot project ID
    :type project_id: str
    :param dataset_id: DataRobot AI Catalog dataset ID
    :type dataset_id: str
    :param credential_id: DataRobot Credentials ID
    :type credential_id: str
    :param dataset_version_id: DataRobot AI Catalog dataset version ID
    :type dataset_version_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: External dataset ID for the model
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "dataset_id",
        "credential_id",
        "dataset_version_id",
    ]

    def __init__(
        self,
        *,
        project_id: str,
        dataset_id: str,
        credential_id: Optional[str] = None,
        dataset_version_id: Optional[str] = None,
        max_wait_sec: int = DEFAULT_MAX_WAIT_SEC,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.dataset_id = dataset_id
        self.credential_id = credential_id
        self.dataset_version_id = dataset_version_id
        self.max_wait_sec = max_wait_sec

    def validate(self) -> None:
        if self.project_id is None:
            raise ValueError("project_id is required to add external dataset.")

        if self.dataset_id is None:
            raise ValueError("dataset_id is required to add external dataset.")

    def execute(self, context: Context) -> str:
        project = dr.Project.get(self.project_id)

        external_dataset = project.upload_dataset_from_catalog(
            dataset_id=self.dataset_id,
            credential_id=self.credential_id,
            dataset_version_id=self.dataset_version_id,
            max_wait=self.max_wait_sec,
        )

        self.log.info(
            f"External Dataset added to the Project, external dataset_id={external_dataset.id}"
        )

        return external_dataset.id


class RequestModelPredictionsOperator(BaseDatarobotOperator):
    """
    Requests predictions against a previously uploaded dataset.
    :param project_id: DataRobot project ID
    :type project_id: str
    :param model_id: DataRobot model ID
    :type model_id: str
    :param external_dataset_id: DataRobot external dataset ID
    :type external_dataset_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: Model predictions job ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "model_id",
        "external_dataset_id",
    ]

    def __init__(
        self,
        *,
        project_id: str,
        model_id: str,
        external_dataset_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.model_id = model_id
        self.external_dataset_id = external_dataset_id

    def validate(self) -> None:
        if self.project_id is None:
            raise ValueError("project_id is required to compute model predictions.")

        if self.model_id is None:
            raise ValueError("model_id is required to compute model predictions.")

        if self.external_dataset_id is None:
            raise ValueError("external_dataset_id is required to compute model predictions.")

    def execute(self, context: Context) -> str:
        model = dr.models.Model.get(self.project_id, self.model_id)

        predict_job = model.request_predictions(dataset_id=self.external_dataset_id)

        self.log.info(f"Model predictions requested, job_id={predict_job.id}")

        return predict_job.id



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

from airflow.decorators import dag
from datarobot import AUTOPILOT_MODE

from datarobot_provider.operators.autopilot import StartAutopilotOperator
from datarobot_provider.operators.data_prep import CreateFeatureDiscoveryRecipeOperator
from datarobot_provider.operators.data_registry import UploadDatasetOperator
from datarobot_provider.operators.datarobot import CreateProjectOperator
from datarobot_provider.operators.datarobot import GetOrCreateUseCaseOperator
from datarobot_provider.operators.datarobot import SelectBestModelOperator
from datarobot_provider.operators.deployment import ReplaceModelOperator
from datarobot_provider.operators.model_registry import CreateRegisteredModelVersionOperator
from datarobot_provider.sensors.datarobot import AutopilotCompleteSensor


@dag(
    schedule=None,
    tags=["example", "csv", "feature discovery"],
    params={
        "autopilot_settings": {
            "target": "BadLoan",
            "mode": AUTOPILOT_MODE.QUICK,
        },
        "advanced_options": {"feature_discovery_supervised_feature_reduction": True},
        "deployment_id": "",
    },
)
def datarobot_feature_discovery_retraining_and_scoring(
    primary_dataset_path=(
        "https://s3.us-east-1.amazonaws.com/datarobot_public_datasets/LendingClub/target.csv"
    ),
    profile_dataset_path=(
        "https://s3.us-east-1.amazonaws.com/datarobot_public_datasets/LendingClub/profile.csv"
    ),
    transactions_dataset_path=(
        "https://s3.us-east-1.amazonaws.com/datarobot_public_datasets/LendingClub/transactions.csv"
    ),
):
    # Create a Use Case to keep all subsequent assets. Default name is "Airflow"
    create_use_case = GetOrCreateUseCaseOperator(task_id="create_use_case")

    # Upload all the datasets into Data Registry.
    upload_primary_dataset = UploadDatasetOperator(
        task_id="upload_primary_dataset",
        file_path=primary_dataset_path,
        use_case_id=create_use_case.output,
    )
    upload_profile_dataset = UploadDatasetOperator(
        task_id="upload_profile_dataset",
        file_path=profile_dataset_path,
        use_case_id=create_use_case.output,
    )
    upload_transactions_dataset = UploadDatasetOperator(
        task_id="upload_transactions_dataset",
        file_path=transactions_dataset_path,
        use_case_id=create_use_case.output,
    )

    # Define the secondary datasets for Feature Discovery
    profile_dataset_definition = {
        "identifier": "profile",
        "catalogId": upload_profile_dataset.output,
    }

    transaction_dataset_definition = {
        "identifier": "transactions",
        "catalogId": upload_transactions_dataset.output,
    }

    # Define the relationships between the datasets.
    # You do not need to specify dataset1Identifier when joining with the primary dataset.
    primary_profile_relationship = {
        "dataset2Identifier": "profile",  # to profile
        "dataset1Keys": ["CustomerID"],  # on CustomerID
        "dataset2Keys": ["CustomerID"],
    }

    profile_transaction_relationship = {
        "dataset1Identifier": "profile",  # join profile
        "dataset2Identifier": "transactions",  # to transactions
        "dataset1Keys": ["CustomerID"],  # on CustomerID
        "dataset2Keys": ["CustomerID"],
    }

    dataset_definitions = [profile_dataset_definition, transaction_dataset_definition]
    relationships = [primary_profile_relationship, profile_transaction_relationship]

    feature_discovery_settings = [
        {"name": "enable_days_from_prediction_point", "value": False},
        {"name": "enable_word_count", "value": False},
        {"name": "enable_pairwise_time_difference", "value": True},
        {"name": "enable_numeric_sum", "value": True},
    ]

    # Create a Feature Discovery recipe with all the above information.
    create_feature_discovery_recipe = CreateFeatureDiscoveryRecipeOperator(
        dataset_id=upload_primary_dataset.output,
        use_case_id=create_use_case.output,
        dataset_definitions=dataset_definitions,
        relationships=relationships,
        feature_discovery_settings=feature_discovery_settings,
        task_id="create_feature_discovery_recipe",
    )

    # Create and launch a project from the recipe.
    create_project = CreateProjectOperator(
        recipe_id=create_feature_discovery_recipe.output,
        use_case_id=create_use_case.output,
        task_id="create_project",
    )

    train_models = StartAutopilotOperator(task_id="train_models", project_id=create_project.output)

    autopilot_complete_sensor = AutopilotCompleteSensor(
        task_id="check_autopilot_complete", project_id=create_project.output
    )

    # select best model
    select_best_model = SelectBestModelOperator(
        task_id="select_best_model",
        project_id=create_project.output,
        metric="RMSE",
    )
    # register model
    register_model = CreateRegisteredModelVersionOperator(
        task_id="register_model",
        model_version_params={
            "model_type": "leaderboard",
            "model_id": select_best_model.output,
            "name": "Highest readmitted score test",
            "registered_model_name": "Highest readmitted score test {{ ts }}",
        },
    )
    # replace model
    replace_model = ReplaceModelOperator(
        task_id="replace_model",
        new_registered_model_version_id=register_model.output,
        deployment_id=str("{{ params.deployment_id }}"),
    )

    (
        create_use_case
        >> [upload_primary_dataset, upload_profile_dataset, upload_transactions_dataset]
        >> create_feature_discovery_recipe
        >> create_project
        >> train_models
        >> autopilot_complete_sensor
        >> select_best_model
        >> register_model
        >> replace_model
    )


datarobot_feature_discovery_retraining_and_scoring()



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any
from typing import List
from typing import Optional

import datarobot as dr
from airflow.exceptions import AirflowFailException
from airflow.utils.context import Context
from pandas import DataFrame

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator


class RequestPredictionsOperator(BaseDatarobotOperator):
    """
    Starts predictions generation job for a given model and dataset.

    Args:
        project_id (str): DataRobot project ID.
        model_id (str): DataRobot model ID.
        dataset_id (Optional[str]): DataRobot dataset ID. Defaults to None.
        file_path (Optional[str]): File path to the dataset. Defaults to None.

    Returns:
        str: PredictJob ID.

    Raises:
        AirflowFailException: If neither dataset_id nor file_path is provided.
        AirflowFailException: If both dataset_id and file_path are provided.
    """

    template_fields: List[str] = ["project_id", "model_id", "dataset_id", "file_path"]

    def __init__(
        self,
        *,
        project_id: str,
        model_id: str,
        dataset_id: Optional[str] = None,
        file_path: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.model_id = model_id
        self.dataset_id = dataset_id
        self.file_path = file_path

    def validate(self) -> None:
        if not self.dataset_id and not self.file_path:
            raise AirflowFailException("Either dataset_id or file_path must be provided")
        if self.dataset_id and self.file_path:
            raise AirflowFailException("Both dataset_id and file_path cannot be provided")

    def execute(self, context: Context) -> str:
        model = dr.Model.get(
            project=self.project_id,
            model_id=self.model_id,
        )

        if self.dataset_id:
            self.log.info(
                f"Requesting predictions for model {self.model_id} using dataset {self.dataset_id}"
            )
            predict_job = model.request_predictions(dataset_id=self.dataset_id)
        else:
            self.log.info(
                f"Requesting predictions for model {self.model_id} using file {self.file_path}"
            )
            predict_job = model.request_predictions(file_path=self.file_path)

        self.log.info(f"Model predictions requested, predict_job_id={predict_job.id}")
        return predict_job.id


class SavePredictionsToDatasetOperator(BaseDatarobotOperator):
    """
    Save the predictions from a PredictJob to a dataset.

    NOTE: This operator currently first downloads the predictions to the Airflow worker
    and then uploads them to DataRobot. This may not be suitable for large datasets.

    Args:
        project_id (str): DataRobot project ID.
        predict_job_id (str): PredictJob ID.

    Returns:
        id: Dataset ID.

    Raises:
        AirflowFailException: If predict_job_id or project_id is not provided.
    """

    template_fields: List[str] = ["project_id", "predict_job_id"]

    def __init__(self, *, project_id: str, predict_job_id: str, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.predict_job_id = predict_job_id

    def validate(self) -> None:
        if not self.predict_job_id or not self.project_id:
            raise AirflowFailException("predict_job_id and project_id must be provided")

    def execute(self, context: Context) -> DataFrame:
        predictions = dr.PredictJob.get_predictions(
            project_id=self.project_id,
            predict_job_id=self.predict_job_id,
        )

        self.log.info(f"Predictions retrieved, predict_job_id={self.predict_job_id}")

        project = dr.Project.get(self.project_id)
        predictions_dataset = project.upload_dataset(
            sourcedata=predictions, dataset_filename=f"prediction-results-{self.predict_job_id}.csv"
        )
        return predictions_dataset.id



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime

from airflow.decorators import dag
from airflow.decorators import task

from datarobot_provider.operators.bias_and_fairness import GetBiasAndFairnessSettingsOperator
from datarobot_provider.operators.bias_and_fairness import UpdateBiasAndFairnessSettingsOperator
from datarobot_provider.operators.segment_analysis import GetSegmentAnalysisSettingsOperator
from datarobot_provider.operators.segment_analysis import UpdateSegmentAnalysisSettingsOperator


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example", "mlops"],
    # Default json config example:
    params={
        "segment_analysis_enabled": True,
        "segment_analysis_attributes": ["race", "gender"],
        "protected_features": ["gender"],
        "preferable_target_value": "True",
        "fairness_metric_set": "equalParity",
        "fairness_threshold": 0.1,
    },
)
def deployment_segment_analysis_settings(deployment_id=None):
    if not deployment_id:
        raise ValueError("Invalid or missing `deployment_id` value")
    get_segment_analysis_settings_before_op = GetSegmentAnalysisSettingsOperator(
        task_id="get_segment_analysis_settings_before",
        # you can pass deployment_id from previous operator here:
        deployment_id=deployment_id,
    )

    get_bias_and_fairness_settings_before_op = GetBiasAndFairnessSettingsOperator(
        task_id="get_bias_and_fairness_settings_before",
        # you can pass deployment_id from previous operator here:
        deployment_id=deployment_id,
    )

    update_segment_analysis_settings_op = UpdateSegmentAnalysisSettingsOperator(
        task_id="update_segment_analysis_settings",
        # you can pass deployment_id from previous operator here:
        deployment_id=deployment_id,
    )

    update_bias_and_fairness_settings_op = UpdateBiasAndFairnessSettingsOperator(
        task_id="update_bias_and_fairness_settings",
        # you can pass deployment_id from previous operator here:
        deployment_id=deployment_id,
    )

    get_segment_analysis_settings_after_op = GetSegmentAnalysisSettingsOperator(
        task_id="get_segment_analysis_after",
        # you can pass deployment_id from previous operator here:
        deployment_id=deployment_id,
    )

    get_bias_and_fairness_settings_after_op = GetBiasAndFairnessSettingsOperator(
        task_id="get_bias_and_fairness_settings_after",
        # you can pass deployment_id from previous operator here:
        deployment_id=deployment_id,
    )

    @task(task_id="example_processing_python")
    def settings_processing(
        segment_analysis_settings_before,
        segment_analysis_settings_after,
        bias_and_fairness_settings_before,
        bias_and_fairness_settings_after,
    ):
        """Example of custom logic based on segment analysis settings from the deployment."""

        # Put your service stat processing logic here:

        segment_analysis_settings_changed = {
            setting: segment_analysis_settings_after[setting]
            for setting in segment_analysis_settings_after
            if segment_analysis_settings_after[setting] != segment_analysis_settings_before[setting]
        }

        bias_and_fairness_settings_changed = {
            setting: bias_and_fairness_settings_after[setting]
            for setting in bias_and_fairness_settings_after
            if bias_and_fairness_settings_after[setting]
            != bias_and_fairness_settings_before[setting]
        }

        return segment_analysis_settings_changed, bias_and_fairness_settings_changed

    example_service_stat_processing = settings_processing(
        segment_analysis_settings_before=get_segment_analysis_settings_before_op.output,
        segment_analysis_settings_after=get_segment_analysis_settings_after_op.output,
        bias_and_fairness_settings_before=get_bias_and_fairness_settings_before_op.output,
        bias_and_fairness_settings_after=get_bias_and_fairness_settings_after_op.output,
    )

    (
        get_segment_analysis_settings_before_op
        >> get_bias_and_fairness_settings_before_op
        >> update_segment_analysis_settings_op
        >> update_bias_and_fairness_settings_op
        >> get_segment_analysis_settings_after_op
        >> get_bias_and_fairness_settings_after_op
        >> example_service_stat_processing
    )


deployment_segment_analysis_dag = deployment_segment_analysis_settings()

if __name__ == "__main__":
    deployment_segment_analysis_dag.test()



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections.abc import Sequence
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import datarobot as dr
from airflow.utils.context import Context
from datarobot.models import ModelParameters

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator


class TrainModelOperator(BaseDatarobotOperator):
    """
    Submit a job to the queue to train a model from a specific blueprint.

    Args:
        project_id (str): DataRobot project ID.
        blueprint_id (str): DataRobot blueprint ID.
        featurelist_id (str, optional): The identifier of the featurelist to use. If not defined,
            the default for this project is used.
        source_project_id (str, optional): Which project created this blueprint_id. If None, it defaults to
            looking in this project. Note that you must have read permissions in this project.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.

    Returns:
        str: Model training job ID.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "blueprint_id",
        "featurelist_id",
        "source_project_id",
    ]

    def __init__(
        self,
        *,
        project_id: str,
        blueprint_id: str,
        featurelist_id: Optional[str] = None,
        source_project_id: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.blueprint_id = blueprint_id
        self.featurelist_id = featurelist_id
        self.source_project_id = source_project_id

    def validate(self) -> None:
        if not self.project_id:
            raise ValueError("project_id is required.")

        if not self.blueprint_id:
            raise ValueError("blueprint_id is required.")

    def execute(self, context: Context) -> str:
        project: dr.Project = dr.Project.get(self.project_id)
        blueprint = dr.Blueprint.get(self.project_id, self.blueprint_id)

        job_id = project.train(
            blueprint,
            sample_pct=context["params"].get("sample_pct", None),
            featurelist_id=self.featurelist_id,
            source_project_id=self.source_project_id,
            scoring_type=context["params"].get("scoring_type", None),
            training_row_count=context["params"].get("training_row_count", None),
            n_clusters=context["params"].get("n_clusters", None),
        )

        self.log.info(f"Model Training Job submitted job_id={job_id}")

        return job_id


class RetrainModelOperator(BaseDatarobotOperator):
    """
    Submit a job to the queue to retrain a model on a specific sample size and/or custom featurelist.

    Args:
        project_id (str): DataRobot project ID.
        model_id (str): DataRobot model ID.
        featurelist_id (str, optional): The identifier of the featurelist to use. If not defined, the default
            for this project is used.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.

    Returns:
        str: Model retraining job ID.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "model_id",
        "featurelist_id",
    ]

    def __init__(
        self,
        *,
        project_id: str,
        model_id: str,
        featurelist_id: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.model_id = model_id
        self.featurelist_id = featurelist_id

    def validate(self) -> None:
        if self.project_id is None:
            raise ValueError("project_id is required.")

        if self.model_id is None:
            raise ValueError("model_id is required.")

    def execute(self, context: Context) -> str:
        model = dr.Model.get(self.project_id, self.model_id)

        job_id = model.train(
            featurelist_id=self.featurelist_id,
            sample_pct=context["params"].get("sample_pct", None),
            scoring_type=context["params"].get("scoring_type", None),
            training_row_count=context["params"].get("training_row_count", None),
        )

        self.log.info(f"Model Retraining Job submitted job_id={job_id}")

        return job_id


class AdvancedTuneModelOperator(BaseDatarobotOperator):
    """
    Advanced tune a model using a set of parameters.

    Args:
        project_id (str): DataRobot project ID.
        model_id (str): DataRobot model ID.
        parameters (List[Tuple[str, str, Any]]): List of tuples containing the task name, parameter name, and value.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.

    Returns:
        str: Model retraining job ID.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["project_id", "model_id", "parameters"]

    def __init__(
        self,
        *,
        project_id: str,
        model_id: str,
        parameters: List[Tuple[str, str, Any]],
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.model_id = model_id
        self.parameters = parameters

    def validate(self) -> None:
        if self.project_id is None:
            raise ValueError("project_id is required.")

        if self.model_id is None:
            raise ValueError("model_id is required.")

        if self.parameters is None:
            raise ValueError("parameters is required.")

    def execute(self, context: Context) -> str:
        model = dr.Model.get(self.project_id, self.model_id)
        tune = model.start_advanced_tuning_session()

        for task_name, parameter_name, value in self.parameters:
            tune.set_parameter(task_name=task_name, parameter_name=parameter_name, value=value)

        job = tune.run()
        return job.id


class GetTrainedModelParametersOperator(BaseDatarobotOperator):
    """
    Retrieve the parameters used to train a given model ID.

    Args:
        project_id (str): DataRobot project ID.
        model_id (str): DataRobot model ID.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.

    Returns:
        str: Model retraining job ID.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["project_id", "model_id"]

    def __init__(
        self,
        *,
        project_id: str,
        model_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.model_id = model_id

    def validate(self) -> None:
        if self.project_id is None:
            raise ValueError("project_id is required.")

        if self.model_id is None:
            raise ValueError("model_id is required.")

    def execute(self, context: Context) -> Dict[str, Any]:
        model = dr.Model.get(self.project_id, self.model_id)
        model_parameters: ModelParameters = model.get_parameters()
        return model_parameters.parameters


class CrossValidateModelOperator(BaseDatarobotOperator):
    """
    By default, models are evaluated on the first validation partition. This will compute scoring for all
    cross validation partitions of the model.

    For datetime partitioned projects, see the ScoreBacktestsModelOperator.

    Args:
        project_id (str): DataRobot project ID.
        model_id (str): DataRobot model ID.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.

    Returns:
        str: Cross validation job ID.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "model_id",
    ]

    def __init__(
        self,
        *,
        project_id: str,
        model_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.model_id = model_id

    def validate(self) -> None:
        if not self.project_id:
            raise ValueError("project_id is required.")

        if not self.model_id:
            raise ValueError("blueprint_id is required.")

    def execute(self, context: Context) -> str:
        model = dr.Model.get(self.project_id, self.model_id)
        job = model.cross_validate()

        return job.id


class ScoreBacktestsModelOperator(CrossValidateModelOperator):
    """
    For a Datetime Partitioned Project, backtesting is the only cross-validation method supported.
    For non-datetime partitioned projects, see the CrossValidateModelOperator.

    By default, models are evaluated on the first backtest only. This will compute scoring for all
    backtest partitions of the model.

    Args:
        project_id (str): DataRobot project ID.
        model_id (str): DataRobot model ID.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.

    Returns:
        str: Score backtests job ID.
    """

    def __init__(
        self,
        *,
        project_id: str,
        model_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(project_id=project_id, model_id=model_id, **kwargs)

    def execute(self, context: Context) -> str:
        model = dr.DatetimeModel.get(self.project_id, self.model_id)
        job = model.score_backtests()

        return job.id



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections.abc import Sequence
from typing import Any
from typing import Optional

import datarobot as dr
from airflow.utils.context import Context
from datarobot import BatchMonitoringJob

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator


class BatchMonitoringOperator(BaseDatarobotOperator):
    """
    Creates a batch monitoring job in DataRobot.
    :param deployment_id: DataRobot deployment ID
    :type deployment_id: str
    :param datastore_id: DataRobot DataStore ID for jdbc source connection
    :type datastore_id: str
    :param credential_id: DataRobot Credentials ID for source connection
    :type credential_id: str
    :param datarobot_conn_id: Connection ID, defaults to `datarobot_default`
    :type datarobot_conn_id: str, optional
    :return: Batch Monitoring job ID
    :rtype: str
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "deployment_id",
        "datastore_id",
        "credential_id",
    ]

    def __init__(
        self,
        *,
        deployment_id: Optional[str] = None,
        datastore_id: Optional[str] = None,
        credential_id: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.deployment_id = deployment_id
        self.datastore_id = datastore_id
        self.credential_id = credential_id

    def execute(self, context: Context) -> str:
        monitoring_job_settings = context["params"]["monitoring_settings"]

        # in case of deployment_id was not set from operator argument:
        if self.deployment_id is None:
            self.deployment_id = context["params"]["deployment_id"]

        if self.credential_id is not None:
            monitoring_job_settings["intake_settings"]["credential_id"] = self.credential_id

        intake_settings = monitoring_job_settings.get("intake_settings", dict())

        intake_type = intake_settings.get("type")

        # in case of JDBC intake from operator argument:
        if intake_type == "jdbc" and self.datastore_id is not None:
            monitoring_job_settings["intake_settings"]["data_store_id"] = self.datastore_id

        self.log.info(
            f"Loading monitoring data for deployment_id={self.deployment_id} "
            f"with settings: {monitoring_job_settings}"
        )

        if intake_type == "dataset":
            dataset_id = intake_settings.get("dataset_id")
            if not dataset_id:
                raise ValueError(
                    "Invalid or missing `dataset_id` value for the `dataset` intake type."
                )
            dataset = dr.Dataset.get(dataset_id)
            intake_settings["dataset"] = dataset

            # We no longer need the ID
            del intake_settings["dataset_id"]

        job = BatchMonitoringJob.run(self.deployment_id, **monitoring_job_settings)

        self.log.info(f"Monitoring Job submitted job_id={job.id}")

        return job.id



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import argparse
import importlib
from collections import defaultdict
from pathlib import Path

import black
import yaml
from numpydoc.docscrape import NumpyDocString


class GenerateOperators:
    def __init__(self, whitelist):
        self.whitelist = whitelist

    def generate(self):
        """Generate operator python code for each class/method in the whitelist."""
        generated_code = {}
        for module_key in self.whitelist.keys():
            module = importlib.import_module(module_key)
            for module_obj_key, methods in self.whitelist[module_key].items():
                module_obj = getattr(module, module_obj_key)
                for method in methods:
                    operator_name = f"{module_obj_key}{method.title()}Operator"
                    method_docstring = getattr(module_obj, method).__doc__
                    op_python_code = self.generate_operator_for_method(
                        module_key, module_obj_key, method, method_docstring
                    )
                    generated_code[f"operators.{module_obj_key.lower()}.{operator_name}"] = (
                        op_python_code
                    )
        return generated_code

    @classmethod
    def generate_operator_for_method(
        cls, module_key: str, module_obj_key: str, method: str, method_docstring: NumpyDocString
    ) -> str:
        """Generate the operator code for a given method in a module.

        Parameters
        ----------
        module_key : str
            The module name.
        module_obj_key : str
            The object name within the module.
        method : str
            The method name.
        method_docstring : NumpyDocString
            The docstring for the method.

        Returns
        -------
        The generated operator code.
        """
        print(f"Generating operator for {module_obj_key} {method}")
        doc = NumpyDocString(str(method_docstring))
        operator_name = f"{module_obj_key}{method.title()}Operator"
        op_python_code = (
            f"class {operator_name}(BaseOperator):\n"
            + cls.construct_operator_docstring(doc)
            + "\n\n"
            + cls.construct_operator_attibutes(doc)
            + "\n\n"
            + cls.construct_operator_init(doc)
            + "\n\n"
            + cls.construct_operator_execute(doc, module_key, module_obj_key, method)
            + "\n"
        )
        op_python_code = (
            cls.construct_imports(op_python_code, module_key, module_obj_key)
            + "\n\n"
            + op_python_code
        )
        op_python_code = cls.format_code_block_with_black(op_python_code)
        return op_python_code

    @staticmethod
    def construct_imports(code: str, module_key: str, module_obj_key: str) -> str:
        """Construct the imports code for the operator.

        Parameters
        ----------
        module_key : str
            The module name.
        module_obj_key : str
            The object name within the module.

        Returns
        -------
        The constructed imports code.
        """
        imports = [
            ("Sequence", "from collections.abc import Sequence"),
            ("Any", "from typing import Any"),
            ("Optional", "from typing import Optional"),
            ("",),
            ("AirflowException", "from airflow.exceptions import AirflowException"),
            ("BaseOperator", "from airflow.models import BaseOperator"),
            ("Context", "from airflow.utils.context import Context"),
            (f"from {module_key} import {module_obj_key}",),
            ("",),
            ("DATAROBOT_CONN_ID", "from datarobot_provider.constants import DATAROBOT_CONN_ID"),
            ("DataRobotHook", "from datarobot_provider.hooks.datarobot import DataRobotHook"),
        ]
        actual_imports = []
        for imprt in imports:
            # tuples with length one are required imports
            if len(imprt) == 1:
                actual_imports.append(imprt[0])
            # others are optional imports
            else:
                if imprt[0] in code:
                    actual_imports.append(imprt[1])

        return "\n".join(actual_imports)

    @staticmethod
    def format_code_block_with_black(code_block: str) -> str:
        """Reformat a string of python code using black."""
        return str(black.format_file_contents(code_block, fast=False, mode=black.FileMode()))

    @staticmethod
    def construct_operator_docstring(docstring: NumpyDocString) -> str:
        """Construct the operator docstring based on the provided NumpyDocString.

        Parameters
        ----------
        docstring : NumpyDocString
            The NumpyDocString object containing the documentation.

        Returns
        -------
        str
            The constructed operator docstring.
        """
        op_docstring = f'\t"""\t{" ".join(docstring["Summary"])}\n\n'
        for param in docstring["Parameters"]:
            op_docstring += f"\t:param {param.name}: {' '.join(param.desc)}\n"
            op_docstring += f"\t:type {param.name}: {param.type}\n"
        for returns in docstring["Returns"]:
            op_docstring += f"\t:return: {returns.name}\n"
            op_docstring += f"\t:rtype {returns.type}\n"
        op_docstring += '\t"""\n'
        return op_docstring

    @staticmethod
    def construct_operator_attibutes(docstring: NumpyDocString) -> str:
        """Construct the operator attributes code based on the provided NumpyDocString.

        Parameters
        ----------
        docstring : NumpyDocString
            The NumpyDocString object containing the documentation.

        Returns
        -------
        str
            The constructed operator attributes code.
        """
        param_names = [param.name for param in docstring["Parameters"]]
        param_types = [param.type for param in docstring["Parameters"]]
        op_attributes = f"\ttemplate_fields: Sequence[str] = {param_names}\n"
        op_attributes += (
            f"\ttemplate_fields_renderers: dict[{', '.join(map(str, param_types))}] = {{}}\n"
        )
        op_attributes += "\ttemplate_ext: Sequence[str] = ()\n"
        op_attributes += '\tui_color = "#f4a460"\n'
        return op_attributes

    @staticmethod
    def construct_operator_init(docstring: NumpyDocString) -> str:
        """Construct the operator init code based on the provided NumpyDocString.

        Parameters
        ----------
        docstring : NumpyDocString
            The NumpyDocString object containing the documentation.

        Returns
        -------
        str
            The constructed operator init code.
        """
        op_init = "\tdef __init__(self,*,"
        for param in docstring["Parameters"]:
            op_init += f"{param.name}: Optional[{param.type}] = None,"
        op_init += "datarobot_conn_id: str = DATAROBOT_CONN_ID,\n"
        op_init += "**kwargs: Any) -> None:\n"
        op_init += "\t\tsuper().__init__(**kwargs)\n"
        for param in docstring["Parameters"]:
            op_init += f"\t\tself.{param.name} = {param.name}\n"
        op_init += "\t\tself.datarobot_conn_id = datarobot_conn_id\n"
        return op_init

    @staticmethod
    def construct_operator_execute(
        docstring: NumpyDocString,
        operator_module: str,
        operator_name: str,
        operator_method: str,
    ) -> str:
        """Construct the operator execute code based on the provided NumpyDocString.

        Parameters
        ----------
        docstring : NumpyDocString
            The NumpyDocString object containing the documentation.
        operator_module : str
            The module name.
        operator_name : str
            The name of the operator.
        operator_method : str
            The method name.

        Returns
        -------
        str
            The constructed operator execute code.
        """
        op_execute = "\tdef execute(self, context: Context) -> str:\n"
        op_execute += "\t\t# Initialize DataRobot client\n"
        op_execute += "\t\tDataRobotHook(datarobot_conn_id=self.datarobot_conn_id).run()\n\n"
        for param in docstring["Parameters"]:
            op_execute += f"\t\tif self.{param.name} is None:\n"
            op_execute += (
                f'\t\t\traise ValueError("{param.name} is required for {operator_name}.")\n'
            )
        op_execute += "\n"
        op_execute += f"\t\tresult = {operator_name}.{operator_method}(\n\t\t"
        for param in docstring["Parameters"]:
            op_execute += f"self.{param.name}, "
        op_execute += "\n\t)\n"
        op_execute += "\t\treturn result.id\n"
        return op_execute


parser = argparse.ArgumentParser(description="Generate code for DataRobot provider")
parser.add_argument("--whitelist", type=str, help="Whitelist file path")
parser.add_argument("--output_folder", type=str, help="Output folder path")


def to_snake_case(camel_case_str: str) -> str:
    return "".join(["_" + i.lower() if i.isupper() else i for i in camel_case_str]).lstrip("_")


def main(whitelist_path: Path, output_folder: Path) -> None:
    # load whitelist yaml
    with open(whitelist_path, "r") as f:
        whitelist = yaml.safe_load(f)

    # generate operators
    generator = GenerateOperators(whitelist)
    generated_code = generator.generate()
    modules: dict[str, dict[str, str]] = defaultdict(dict)
    for operator_key, code in generated_code.items():
        _, module_name, operator_name = operator_key.split(".")
        modules[module_name][operator_name] = code

    # write generated code to files
    for module_name, operators in modules.items():
        module_folder = output_folder / module_name
        module_folder.mkdir(parents=True, exist_ok=True)
        for operator_name, code in operators.items():
            # transform operator name from camel case to snake case
            operator_file_name = to_snake_case(operator_name)
            with open(module_folder / f"{operator_file_name}.py", "w") as f:
                f.write(code)

        # write init file, one line for each module
        with open(module_folder / "__init__.py", "w") as f:
            for operator_name in operators.keys():
                operator_file_name = to_snake_case(operator_name)
                f.write(f"from .{operator_file_name} import {operator_name} as {operator_name}\n")


if __name__ == "__main__":
    args = parser.parse_args()
    main(Path(args.whitelist), Path(args.output_folder))



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""
Config example for this dag:
{
    "dataset_file_path": "/tests/integration/datasets/titanic.csv",
}
"""

from datetime import datetime

from airflow.decorators import dag

from datarobot_provider.operators.data_registry import UploadDatasetOperator


@dag(
    schedule=None,
    start_date=datetime(2022, 1, 1),
    tags=["example"],
    params={"dataset_file_path": "./titanic.csv"},
)
def datarobot_dataset_uploading():
    dataset_uploading_op = UploadDatasetOperator(
        task_id="dataset_uploading",
    )

    dataset_uploading_op


datarobot_pipeline_dag = datarobot_dataset_uploading()

if __name__ == "__main__":
    print(datarobot_pipeline_dag.test())



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime

from airflow.decorators import dag
from airflow.operators.python import BranchPythonOperator

from datarobot_provider.operators.deployment import ActivateDeploymentOperator
from datarobot_provider.operators.deployment import GetDeploymentStatusOperator


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example", "mlops"],
)
def deployment_deactivate_activate(deployment_id=None):
    if not deployment_id:
        raise ValueError("Invalid or missing `deployment_id` value")

    get_deployment_status_initial_op = GetDeploymentStatusOperator(
        task_id="get_deployment_status_initial",
        deployment_id=deployment_id,
    )

    def choose_branch(deployment_status):
        if deployment_status == "inactive":
            return ["activate_deployment"]
        return ["deactivate_deployment"]

    branching = BranchPythonOperator(
        task_id="branching",
        python_callable=choose_branch,
        op_args=[get_deployment_status_initial_op.output],
    )

    deactivate_deployment_op = ActivateDeploymentOperator(
        task_id="deactivate_deployment",
        activate=False,
        deployment_id=deployment_id,
    )

    activate_deployment_op = ActivateDeploymentOperator(
        task_id="activate_deployment",
        activate=True,
        deployment_id=deployment_id,
    )

    (
        get_deployment_status_initial_op
        >> branching
        >> (activate_deployment_op, deactivate_deployment_op)
    )


deployment_deactivate_activate_dag = deployment_deactivate_activate()

if __name__ == "__main__":
    deployment_deactivate_activate_dag.test()



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any

import datarobot as dr
from airflow.sensors.base import BaseSensorOperator
from airflow.utils.context import Context
from datarobot.errors import AsyncFailureError
from datarobot.errors import AsyncProcessUnsuccessfulError
from datarobot.models import StatusCheckJob

from datarobot_provider.constants import DATAROBOT_CONN_ID
from datarobot_provider.hooks.datarobot import DataRobotHook


class AutopilotCompleteSensor(BaseSensorOperator):
    """
    Checks if the DataRobot Autopilot has completed training the models.

    Args:
        project_id (str): DataRobot project ID.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.

    Returns:
        bool: False if not yet completed, True when complete.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields = ["project_id"]

    def __init__(
        self,
        *,
        project_id: str,
        datarobot_conn_id: str = DATAROBOT_CONN_ID,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.datarobot_conn_id = datarobot_conn_id

        self.hook = DataRobotHook(datarobot_conn_id)

    def poke(self, context: Context) -> bool:
        # Initialize DataRobot client
        DataRobotHook(datarobot_conn_id=self.datarobot_conn_id).run()

        self.log.info("Checking if Autopilot is complete")
        project = dr.Project.get(self.project_id)
        if project._autopilot_status_check()["autopilot_done"]:
            return True
        return False


class ScoringCompleteSensor(BaseSensorOperator):
    """
    Checks whether scoring predictions is complete.

    Args:
        job_id (str): Batch prediction job ID.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.

    Returns:
        bool: False if not yet completed, True when complete.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields = ["job_id"]

    def __init__(
        self,
        *,
        job_id: str,
        datarobot_conn_id: str = DATAROBOT_CONN_ID,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.job_id = job_id
        self.datarobot_conn_id = datarobot_conn_id

        self.hook = DataRobotHook(datarobot_conn_id)

    def poke(self, context: Context) -> bool:
        # Initialize DataRobot client
        DataRobotHook(datarobot_conn_id=self.datarobot_conn_id).run()

        self.log.info("Checking if scoring is complete")
        job = dr.BatchPredictionJob.get(self.job_id)
        job_data = job.get_status()
        if job_data["status"].lower()[:5] in ["error", "abort"]:
            raise AsyncProcessUnsuccessfulError(
                f"The job did not complete successfully. Job Data: {job_data}"
            )
        if job_data["status"].lower() == "completed":
            return True
        return False


class StatusCheckJobCompleteSensor(BaseSensorOperator):
    """
    Checks whether a status job is complete.

    Args:
        job_id (str): Job ID.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.

    Returns:
        bool: False if not yet completed, True when complete.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields = ["job_id"]

    def __init__(
        self,
        *,
        job_id: str,
        datarobot_conn_id: str = DATAROBOT_CONN_ID,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.job_id = job_id
        self.datarobot_conn_id = datarobot_conn_id

        self.hook = DataRobotHook(datarobot_conn_id)

    def poke(self, context: Context) -> bool:
        # Initialize DataRobot client
        DataRobotHook(datarobot_conn_id=self.datarobot_conn_id).run()

        self.log.info("Checking if async job is complete")
        job = StatusCheckJob.from_id(self.job_id)
        try:
            job_data = job.get_status()
            if str(job_data.status).lower() in ["error", "abort"]:
                raise AsyncProcessUnsuccessfulError(
                    f"The job did not complete successfully. Job Data: {job_data}"
                )
            if str(job_data.status).lower() == "completed":
                return True
        except AsyncFailureError as err:
            raise AsyncProcessUnsuccessfulError(
                f"The job did not complete successfully. Job ID: {self.job_id}"
            ) from err
        return False



#
# Copyright 2021-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import re

from setuptools import find_packages
from setuptools import setup

common_setup_kwargs = dict(
    name=None,
    version=None,
    description="This client library is designed to integrate Airflow with the DataRobot API.",
    author="datarobot",
    author_email="api-maintainer@datarobot.com",
    maintainer="datarobot",
    maintainer_email="api-maintainer@datarobot.com",
    url="https://datarobot.com",
    project_urls={
        "Documentation": "https://datarobot-datarobot-airflow-provider.readthedocs-hosted.com/en/latest/",
        "Changelog": "https://github.com/datarobot/airflow-provider-datarobot/blob/main/CHANGES.md",
        "Bug Tracker": "https://github.com/datarobot/airflow-provider-datarobot/issues",
        "Source Code": "https://github.com/datarobot/airflow-provider-datarobot",
    },
    license="DataRobot Tool and Utility Agreement",
    packages=None,
    package_data={"airflow_provider_datarobot": ["py.typed"]},
    python_requires=">=3.9",
    long_description=None,
    long_description_content_type="text/markdown",
    classifiers=None,
    install_requires=["apache-airflow>=2.6.0", "datarobot>=3.6.2"],
    entry_points={
        "apache_airflow_provider": [
            "provider_info = datarobot_provider.__init__:get_provider_info",
        ],
    },
    extras_require={
        "dev": [
            "pre-commit>=4.0.1",
            "ruff>=0.9.2",
            "mypy>=0.931",
            "pytest>=7.0.0",
            "pytest-mock>=3.7.0",
            "pytest-helpers-namespace>=2021.12.29",
            "numpydoc>=1.7.0,<1.8.0",
            "black==24.10.0",
            "pyyaml>=6.0.2",
            "types-PyYAML>=6.0.12",
            "freezegun>=1.5.1",
        ],
        "docs": [
            "Sphinx==8.1.3",
            "sphinx_rtd_theme>=3.0",
            "sphinx-autodoc-typehints>=2",
            "pyenchant==3.2.2",
            "sphinx-copybutton",
            "sphinx-markdown-builder",
            "myst-parser==4.0.0",
            "numpydoc>=1.7.0,<1.8.0",
        ],
    },
)

classifiers = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "License :: Other/Proprietary License",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
]

DESCRIPTION_TEMPLATE = """
About {package_name}
=============================================
![package](https://img.shields.io/pypi/v/{package_name}.svg)
![versions](https://img.shields.io/pypi/pyversions/{package_name}.svg)
![status](https://img.shields.io/pypi/status/{package_name}.svg)

This package provides operators, sensors, and a hook to integrate [DataRobot](https://www.datarobot.com)
into Apache Airflow. {extra_desc}

This package is released under the terms of the DataRobot Tool and Utility Agreement, which
can be found on our [Legal](https://www.datarobot.com/legal/) page, along with our privacy policy and more.

Installation
=========================
You must have a datarobot account.

```
$ pip install {pip_package_name}
```

Bug Reporting and Q&A
=========================
To report issues or ask questions, send email to [the team](mailto:api-maintainer@datarobot.com).
"""

with open("datarobot_provider/_version.py") as fd:
    version_search = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE)
    if not version_search:
        raise RuntimeError("Cannot find version information")
    version = version_search.group(1)

if not version:
    raise RuntimeError("Cannot find version information")

# RELEASE SETUP
package_name = "airflow_provider_datarobot"

description = DESCRIPTION_TEMPLATE.format(
    package_name=package_name,
    pypi_url_target="https://pypi.python.org/pypi/airflow-provider-datarobot/",
    extra_desc="",
    pip_package_name=package_name,
)

packages = find_packages(exclude=["docs*", "tests*", "*_experimental*"])

common_setup_kwargs.update(
    name=package_name,
    version=version,
    packages=packages,
    long_description=description,
    classifiers=classifiers,
)

setup(**common_setup_kwargs)



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any

from airflow.exceptions import AirflowException
from airflow.sensors.base import BaseSensorOperator
from airflow.utils.context import Context
from datarobot.errors import AsyncProcessUnsuccessfulError
from datarobot.models.status_check_job import StatusCheckJob

from datarobot_provider.constants import DATAROBOT_CONN_ID
from datarobot_provider.hooks.datarobot import DataRobotHook


class BaseAsyncResolutionSensor(BaseSensorOperator):
    """
    Checks if the DataRobot Async API call has completed.

    Args:
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.

    Returns:
        bool: False if not yet completed.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields = ["job_id"]

    def __init__(
        self,
        *,
        job_id: str,
        datarobot_conn_id: str = DATAROBOT_CONN_ID,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.job_id = job_id
        self.datarobot_conn_id = datarobot_conn_id

        self.hook = DataRobotHook(datarobot_conn_id)

    def poke(self, context: Context) -> bool:
        # Initialize DataRobot client
        DataRobotHook(datarobot_conn_id=self.datarobot_conn_id).run()

        if not self.job_id:
            raise AirflowException("job_id is not defined")

        self.log.info("Checking if DataRobot API async call is complete")

        status_check_job = StatusCheckJob.from_id(self.job_id)
        job_status_result = status_check_job.get_status()

        self.log.debug(f"API async call status:{job_status_result.status}")
        self.log.debug(
            f"API async call completed_resource_url:{job_status_result.completed_resource_url}"
        )

        if job_status_result.status in ["ERROR", "ABORT"]:
            raise AsyncProcessUnsuccessfulError(
                f"The job did not complete successfully. Job Data: {job_status_result.status}"
            )
        elif job_status_result.status == "COMPLETED":
            return True
        return False



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections.abc import Sequence
from typing import Any
from typing import List
from typing import Optional

import datarobot as dr
from airflow.exceptions import AirflowFailException
from airflow.utils.context import Context
from datarobot import DatetimePartitioningSpecification
from datarobot import Project

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator

DATAROBOT_MAX_WAIT = 600


class AutopilotBaseOperator(BaseDatarobotOperator):
    @staticmethod
    def start_autopilot(
        project_id,
        autopilot_settings,
        partitioning_settings=None,
        datetime_partitioning_settings=None,
        advanced_options=None,
        featurelist_id=None,
        relationships_configuration_id=None,
        segmentation_task_id=None,
        max_wait_sec=DATAROBOT_MAX_WAIT,
    ):
        """
        Triggers DataRobot Autopilot to train set of models.

        Args:
            project_id (str): DataRobot project ID.
            autopilot_settings (dict): Autopilot settings.
            partitioning_settings (dict, optional): Partitioning settings.
            datetime_partitioning_settings (dict, optional): Datetime partitioning settings.
            advanced_options (dict, optional): Advanced options.
            featurelist_id (str, optional): Specifies which feature list to use.
            relationships_configuration_id (str, optional): ID of the relationships configuration to use.
            segmentation_task_id (str, optional): The segmentation task that should be used to split the project for
                segmented modeling.
            max_wait_sec (int, optional): For some settings, an asynchronous task must be run to analyze the dataset.
                max_wait governs the maximum time (in seconds) to wait before giving up.
            datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.
        """

        project = dr.Project.get(project_id)
        if project.target:
            raise AirflowFailException(f"Autopilot has already started for project_id={project.id}")

        if featurelist_id:
            autopilot_settings["featurelist_id"] = featurelist_id

        if relationships_configuration_id:
            autopilot_settings["relationships_configuration_id"] = relationships_configuration_id

        if segmentation_task_id:
            autopilot_settings["segmentation_task_id"] = segmentation_task_id

        autopilot_settings["max_wait"] = max_wait_sec

        if datetime_partitioning_settings and partitioning_settings:
            raise AirflowFailException(
                "parameters: datetime_partitioning_settings and partitioning_settings are mutually exclusive"
            )

        if datetime_partitioning_settings:
            if isinstance(datetime_partitioning_settings, dict):
                project.set_datetime_partitioning(**datetime_partitioning_settings)
            else:
                project.set_datetime_partitioning(
                    datetime_partition_spec=datetime_partitioning_settings
                )
        elif partitioning_settings:
            project.set_partitioning_method(**partitioning_settings)

        if advanced_options:
            project.set_options(**advanced_options)

        # finalize the project and start the autopilot
        project.analyze_and_model(**autopilot_settings)


class StartAutopilotOperator(AutopilotBaseOperator):
    """
    Triggers DataRobot Autopilot to train set of models.

    Args:
        project_id (str): DataRobot project ID.
        featurelist_id (str, optional): Specifies which feature list to use.
        relationships_configuration_id (str, optional): ID of the relationships configuration to use.
        segmentation_task_id (str, optional): The segmentation task that should be used to split the project for
            segmented modeling.
        max_wait_sec (int, optional): For some settings, an asynchronous task must be run to analyze the dataset.
            max_wait governs the maximum time (in seconds) to wait before giving up.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "featurelist_id",
        "relationships_configuration_id",
        "segmentation_task_id",
    ]

    def __init__(
        self,
        *,
        project_id: str,
        featurelist_id: Optional[str] = None,
        relationships_configuration_id: Optional[str] = None,
        segmentation_task_id: Optional[str] = None,
        max_wait_sec: int = DATAROBOT_MAX_WAIT,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.featurelist_id = featurelist_id
        self.relationships_configuration_id = relationships_configuration_id
        self.segmentation_task_id = segmentation_task_id
        self.max_wait_sec = max_wait_sec

    def execute(self, context: Context) -> None:
        self.log.info(f"Starting DataRobot Autopilot for project_id={self.project_id}")
        self.start_autopilot(
            project_id=self.project_id,
            autopilot_settings=context["params"]["autopilot_settings"],
            partitioning_settings=context["params"].get("partitioning_settings"),
            datetime_partitioning_settings=context["params"].get("datetime_partitioning_settings"),
            advanced_options=context["params"].get("advanced_options"),
            featurelist_id=self.featurelist_id,
            relationships_configuration_id=self.relationships_configuration_id,
            segmentation_task_id=self.segmentation_task_id,
            max_wait_sec=DATAROBOT_MAX_WAIT,
        )


class StartDatetimeAutopilotOperator(AutopilotBaseOperator):
    """
    Configure the project as a datetime partitioned project. This operator should be run after
    the CreateProjectOperator and before training models. It will return a

    Args:
        project_id (str): DataRobot project ID.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.
        datetime_partition_column (str): The name of the column whose values as dates are used to assign a
            row to a particular partition.
        featurelist_id (str, optional): Specifies which feature list to use.
        segmentation_task_id (str, optional): The segmentation task that should be used to split the project for
            segmented modeling.
        autopilot_data_selection_method (str): One of `datarobot.enums.DATETIME_AUTOPILOT_DATA_SELECTION_METHOD`.
            Whether models created by the autopilot should use "rowCount" or "duration" as their data_selection_method.
        validation_duration (str or None): The default validation_duration for the backtests.
        holdout_start_date (datetime.datetime or None): The start date of holdout scoring data. If `holdout_start_date`
            is specified, either `holdout_duration` or `holdout_end_date` must also be specified. If `disable_holdout`
            is set to `True`, `holdout_start_date`, `holdout_duration`, and `holdout_end_date` may not be specified.
        holdout_duration (str or None): The duration of the holdout scoring data. If `holdout_duration` is specified,
            `holdout_start_date` must also be specified. If `disable_holdout` is set to `True`, `holdout_duration`,
            `holdout_start_date`, and `holdout_end_date` may not be specified.
        holdout_end_date (datetime.datetime or None): The end date of holdout scoring data. If `holdout_end_date` is
            specified, `holdout_start_date` must also be specified. If `disable_holdout` is set to `True`,
            `holdout_end_date`, `holdout_start_date`, and `holdout_duration` may not be specified.
        disable_holdout (bool or None): Whether to suppress allocating a holdout fold. If set to `True`,
            `holdout_start_date`, `holdout_duration`, and `holdout_end_date` may not be specified.
        gap_duration (str or None): The duration of the gap between training and holdout scoring data.
        number_of_backtests (int or None): The number of backtests to use.
        backtests (list of datarobot.BacktestSpecification): The exact specification of backtests to use. The indices
            of the specified backtests should range from 0 to number_of_backtests - 1. If any backtest is left
            unspecified, a default configuration will be chosen.
        use_time_series (bool): Whether to create a time series project (if `True`) or an OTV project which uses
            datetime partitioning (if `False`). The default behavior is to create an OTV project.
        default_to_known_in_advance (bool): Optional, default `False`. Used for time series projects only. Sets
            whether all features default to being treated as known in advance. Known in advance features are expected
            to be known for dates in the future when making predictions, e.g., "is this a holiday?". Individual
            features can be set to a value different than the default using the `feature_settings` parameter.
        default_to_do_not_derive (bool): Optional, default `False`. Used for time series projects only. Sets whether
            all features default to being treated as do-not-derive features, excluding them from feature derivation.
            Individual features can be set to a value different than the default by using the `feature_settings`
            parameter.
        feature_derivation_window_start (int or None): Only used for time series projects. Offset into the past to
            define how far back relative to the forecast point the feature derivation window should start. Expressed
            in terms of the `windows_basis_unit` and should be negative value or zero.
        feature_derivation_window_end (int or None): Only used for time series projects. Offset into the past to define
            how far back relative to the forecast point the feature derivation window should end. Expressed in terms of
            the `windows_basis_unit` and should be a negative value or zero.
        feature_settings (list of datarobot.FeatureSettings): Optional, a list specifying per feature settings, can be
            left unspecified.
        forecast_window_start (int or None): Only used for time series projects. Offset into the future to define how
            far forward relative to the forecast point the forecast window should start. Expressed in terms of the
            `windows_basis_unit`.
        forecast_window_end (int or None): Only used for time series projects. Offset into the future to define how far
            forward relative to the forecast point the forecast window should end. Expressed in terms of the
            `windows_basis_unit`.
        windows_basis_unit (str, optional): Only used for time series projects. Indicates which unit is a basis for
        feature derivation window and forecast window. Valid options are detected time unit (one of the
            `datarobot.enums.TIME_UNITS`) or "ROW". If omitted, the default value is the detected time unit.
        treat_as_exponential (str, optional): Defaults to "auto". Used to specify whether to treat data as exponential
            trend and apply transformations like log-transform. Use values from the
            `datarobot.enums.TREAT_AS_EXPONENTIAL` enum.
        differencing_method (str, optional): Defaults to "auto". Used to specify which differencing method to apply of
            case if data is stationary. Use values from `datarobot.enums.DIFFERENCING_METHOD` enum.
        periodicities (list of datarobot.Periodicity, optional): A list of `datarobot.Periodicity`. Periodicities units
            should be "ROW", if the `windows_basis_unit` is "ROW".
        multiseries_id_columns (List[str] or None): A list of the names of multiseries id columns to define series
            within the training data. Currently only one multiseries id column is supported.
        use_cross_series_features (bool): Whether to use cross series features.
        aggregation_type (Optional[str]): The aggregation type to apply when creating cross series features. Optional,
            must be one of "total" or "average".
        cross_series_group_by_columns (list of Optional[str]): List of columns (currently of length 1). Optional setting
            that indicates how to further split series into related groups. For example, if every series is sales of an
            individual product, the series group-by could be the product category with values like "men's clothing",
            "sports equipment", etc. Can only be used in a multiseries project with `use_cross_series_features` set to
            `True`.
        calendar_id (Optional[str]): The id of the `datarobot.CalendarFile` to use with this project.
        unsupervised_mode (Optional[bool]): Defaults to False, indicates whether partitioning should be constructed for
            the unsupervised project.
        model_splits (Optional[int]): Sets the cap on the number of jobs per model used when building models to control
            number of jobs in the queue. Higher number of model splits will allow for less downsampling leading to the
            use of more post-processed data.
        allow_partial_history_time_series_predictions (Optional[bool]): Whether to allow time series models to make
            predictions using partial historical data.
        unsupervised_type (Optional[str]): The unsupervised project type, only valid if `unsupervised_mode` is True. Use
            values from `datarobot.enums.UnsupervisedTypeEnum` enum. If not specified then the project defaults to
            'anomaly' when `unsupervised_mode` is True.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = ["project_id", "datetime_partition_column", "use_time_series"]

    def __init__(
        self,
        *,
        project_id: str,
        datetime_partition_column: str,
        use_time_series: bool,
        featurelist_id: Optional[str] = None,
        segmentation_task_id: Optional[str] = None,
        autopilot_data_selection_method: Optional[str] = None,
        validation_duration: Optional[str] = None,
        holdout_start_date: Optional[str] = None,
        holdout_duration: Optional[str] = None,
        disable_holdout: Optional[bool] = None,
        gap_duration: Optional[str] = None,
        number_of_backtests: Optional[int] = None,
        backtests: Any = None,
        default_to_known_in_advance: bool = False,
        default_to_do_not_derive: bool = False,
        feature_derivation_window_start: Optional[int] = None,
        feature_derivation_window_end: Optional[int] = None,
        feature_settings: Any = None,
        forecast_window_start: Optional[int] = None,
        forecast_window_end: Optional[int] = None,
        windows_basis_unit: Optional[str] = None,
        treat_as_exponential: Optional[str] = None,
        differencing_method: Optional[str] = None,
        periodicities: Any = None,
        multiseries_id_columns: Optional[List[str]] = None,
        use_cross_series_features: Optional[bool] = None,
        aggregation_type: Optional[str] = None,
        cross_series_group_by_columns: Optional[List[str]] = None,
        calendar_id: Optional[str] = None,
        holdout_end_date: Optional[str] = None,
        unsupervised_mode: bool = False,
        model_splits: Optional[int] = None,
        allow_partial_history_time_series_predictions: bool = False,
        unsupervised_type: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.datetime_partition_column = datetime_partition_column
        self.use_time_series = use_time_series
        self.featurelist_id = featurelist_id
        self.segmentation_task_id = segmentation_task_id
        self.autopilot_data_selection_method = autopilot_data_selection_method
        self.validation_duration = validation_duration
        self.holdout_start_date = holdout_start_date
        self.holdout_duration = holdout_duration
        self.disable_holdout = disable_holdout
        self.gap_duration = gap_duration
        self.number_of_backtests = number_of_backtests
        self.backtests = backtests
        self.default_to_known_in_advance = default_to_known_in_advance
        self.default_to_do_not_derive = default_to_do_not_derive
        self.feature_derivation_window_start = feature_derivation_window_start
        self.feature_derivation_window_end = feature_derivation_window_end
        self.feature_settings = feature_settings
        self.forecast_window_start = forecast_window_start
        self.forecast_window_end = forecast_window_end
        self.windows_basis_unit = windows_basis_unit
        self.treat_as_exponential = treat_as_exponential
        self.differencing_method = differencing_method
        self.periodicities = periodicities
        self.multiseries_id_columns = multiseries_id_columns
        self.use_cross_series_features = use_cross_series_features
        self.aggregation_type = aggregation_type
        self.cross_series_group_by_columns = cross_series_group_by_columns
        self.calendar_id = calendar_id
        self.holdout_end_date = holdout_end_date
        self.unsupervised_mode = unsupervised_mode
        self.model_splits = model_splits
        self.allow_partial_history_time_series_predictions = (
            allow_partial_history_time_series_predictions
        )
        self.unsupervised_type = unsupervised_type

    def execute(self, context: Context) -> None:
        project = Project.get(self.project_id)
        if project.target:
            self.log.info(f"Project has already been started for project_id={project.id}")
        else:
            datetime_partition_spec = self.get_datetime_partition_spec()

            self.start_autopilot(
                project_id=self.project_id,
                autopilot_settings=context["params"]["autopilot_settings"],
                partitioning_settings=None,
                datetime_partitioning_settings=datetime_partition_spec,
                advanced_options=context["params"].get("advanced_options"),
                featurelist_id=self.featurelist_id,
                relationships_configuration_id=None,
                segmentation_task_id=self.segmentation_task_id,
                max_wait_sec=DATAROBOT_MAX_WAIT,
            )

    def get_datetime_partition_spec(self):
        return DatetimePartitioningSpecification(
            use_time_series=self.use_time_series,
            datetime_partition_column=self.datetime_partition_column,
            autopilot_data_selection_method=self.autopilot_data_selection_method,
            validation_duration=self.validation_duration,
            holdout_start_date=self.holdout_start_date,
            holdout_duration=self.holdout_duration,
            disable_holdout=self.disable_holdout,
            gap_duration=self.gap_duration,
            number_of_backtests=self.number_of_backtests,
            backtests=self.backtests,
            default_to_known_in_advance=self.default_to_known_in_advance,
            default_to_do_not_derive=self.default_to_do_not_derive,
            feature_derivation_window_start=self.feature_derivation_window_start,
            feature_derivation_window_end=self.feature_derivation_window_end,
            feature_settings=self.feature_settings,
            forecast_window_start=self.forecast_window_start,
            forecast_window_end=self.forecast_window_end,
            windows_basis_unit=self.windows_basis_unit,
            treat_as_exponential=self.treat_as_exponential,
            differencing_method=self.differencing_method,
            periodicities=self.periodicities,
            multiseries_id_columns=self.multiseries_id_columns,
            use_cross_series_features=self.use_cross_series_features,
            aggregation_type=self.aggregation_type,
            cross_series_group_by_columns=self.cross_series_group_by_columns,
            calendar_id=self.calendar_id,
            holdout_end_date=self.holdout_end_date,
            unsupervised_mode=self.unsupervised_mode,
            model_splits=self.model_splits,
            allow_partial_history_time_series_predictions=self.allow_partial_history_time_series_predictions,
            unsupervised_type=self.unsupervised_type,
        )



# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

from logging import Logger

import datarobot as dr
import pandas as pd
from airflow.decorators import dag

from datarobot_provider.operators.custom_function import CustomFunctionOperator

"""
Example of Aiflow DAG to run any custom function in a DAG, in this case a post processing function on predictions.
Configurable parameters for this dag:
* project_id - the project id for the project to use
* predict_job_id - the predict job id for the predictions to post-process
"""


def post_process_predictions(project_id: str, predict_job_id: str, log: Logger) -> str:
    """
    Manipulate predictions after they have been generated.
    Optionally, the function can accept the log parameter to log messages on the Airflow dashboard.
    """

    log.info(f"Fetching predictions for project {project_id} and predict job {predict_job_id}")
    predictions: pd.DataFrame = dr.PredictJob.get_predictions(
        project_id=project_id,
        predict_job_id=predict_job_id,
    )

    # Do some post-processing on the predictions dataframe here
    ...

    project = dr.Project.get(project_id)
    postprocessed_dataset = project.upload_dataset(
        sourcedata=predictions, dataset_filename=f"postprocessed-predictions-{predict_job_id}.csv"
    )

    log.info(f"Post-processed predictions saved to dataset {postprocessed_dataset.id}")

    return postprocessed_dataset.id


@dag(
    schedule=None,
    tags=["example", "custom_function", "post_predictions_formatting"],
    params={
        "project_id": "YOUR_PROJECT_ID",
        "predict_job_id": "YOUR_PREDICT_JOB_ID",
    },
)
def post_predictions_formatting_dag():
    post_process = CustomFunctionOperator(
        task_id="custom_post_process_function",
        # This can be any arbitrary function
        custom_func=post_process_predictions,
        # The parameters must be specified here as a dictionary
        func_params={
            "project_id": "{{ params.project_id }}",
            "predict_job_id": "{{ params.predict_job_id }}",
        },
    )

    (post_process)


# Instantiate the DAG
post_predictions_formatting_dag()



# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

from datetime import datetime

from airflow.decorators import dag
from airflow.operators.empty import EmptyOperator
from airflow.operators.python import BranchPythonOperator
from datarobot import TARGET_TYPE
from datarobot.enums import DEPLOYMENT_IMPORTANCE

from datarobot_provider.operators.custom_models import CreateCustomInferenceModelOperator
from datarobot_provider.operators.custom_models import CreateCustomModelDeploymentOperator
from datarobot_provider.operators.custom_models import CreateCustomModelVersionOperator
from datarobot_provider.operators.custom_models import CreateExecutionEnvironmentOperator
from datarobot_provider.operators.custom_models import CreateExecutionEnvironmentVersionOperator
from datarobot_provider.operators.custom_models import CustomModelTestOperator
from datarobot_provider.operators.custom_models import GetCustomModelTestOverallStatusOperator
from datarobot_provider.operators.data_registry import UploadDatasetOperator


@dag(
    schedule=None,
    start_date=datetime(2023, 1, 1),
    tags=["example", "custom model"],
    params={
        "execution_environment_name": "Execution Env Airflow",
        "execution_environment_description": "Demo Execution Environment created by Airflow provider",
        "programming_language": "python",
        "required_metadata_keys": [],  # example: [{"field_name": "test_key", "display_name": "test_display_name"}],
        "custom_model_description": "This is a custom model created by Airflow",
        "environment_version_description": "created by Airflow provider",
        "custom_model_name": "Airflow Custom model Demo",
        "target_type": TARGET_TYPE.REGRESSION,
        "target_name": "Grade 2014",
        "is_major_update": True,
        "is_training_data_for_versions_permanently_enabled": True,
        "docker_context_path": "/usr/local/airflow/dags/datarobot-user-models/public_dropin_environments/python3_pytorch/",
        "custom_model_folder": "/usr/local/airflow/dags/datarobot-user-models/model_templates/python3_pytorch/",
        "test_dataset_file_path": "/usr/local/airflow/dags/datarobot-user-models/tests/testdata/juniors_3_year_stats_regression.csv",
        "train_dataset_file_path": "/usr/local/airflow/dags/datarobot-user-models/tests/testdata/juniors_3_year_stats_regression.csv",
    },
)
def create_custom_model_pipeline(
    prediction_server_id="5fbc1924ccfc5a0025c424bf", deployment_name="Demo Deployment Airflow"
):
    create_execution_environment_op = CreateExecutionEnvironmentOperator(
        task_id="create_execution_environment",
    )

    create_execution_environment_version_op = CreateExecutionEnvironmentVersionOperator(
        task_id="create_execution_environment_version",
        execution_environment_id=create_execution_environment_op.output,
    )

    create_custom_inference_model_op = CreateCustomInferenceModelOperator(
        task_id="create_custom_inference_model",
    )

    train_dataset_uploading_op = UploadDatasetOperator(
        task_id="train_dataset_uploading", file_path_param="train_dataset_file_path"
    )

    create_custom_model_version_op = CreateCustomModelVersionOperator(
        task_id="create_custom_model_version",
        custom_model_id=create_custom_inference_model_op.output,
        base_environment_id=create_execution_environment_op.output,
        training_dataset_id=train_dataset_uploading_op.output,
    )

    test_dataset_uploading_op = UploadDatasetOperator(
        task_id="test_dataset_uploading", file_path_param="test_dataset_file_path"
    )

    custom_model_test_op = CustomModelTestOperator(
        task_id="custom_model_test",
        custom_model_id=create_custom_inference_model_op.output,
        custom_model_version_id=create_custom_model_version_op.output,
        dataset_id=test_dataset_uploading_op.output,
    )

    custom_model_test_overall_status_op = GetCustomModelTestOverallStatusOperator(
        task_id="custom_model_test_overall_status",
        custom_model_test_id=custom_model_test_op.output,
    )

    def choose_branch(custom_model_test_overall_status):
        if custom_model_test_overall_status == "succeeded":
            return ["deploy_custom_model"]
        return ["custom_model_test_fail"]

    if_tests_succeed = BranchPythonOperator(
        task_id="if_tests_succeed",
        python_callable=choose_branch,
        op_args=[custom_model_test_overall_status_op.output],
    )

    deploy_custom_model_op = CreateCustomModelDeploymentOperator(
        task_id="deploy_custom_model",
        custom_model_version_id=create_custom_model_version_op.output,
        deployment_name=deployment_name,
        prediction_server_id=prediction_server_id,
        importance=DEPLOYMENT_IMPORTANCE.LOW,
    )

    custom_model_test_fail = EmptyOperator(
        task_id="custom_model_test_fail",
    )

    (
        create_execution_environment_op
        >> create_execution_environment_version_op
        >> create_custom_inference_model_op
        >> train_dataset_uploading_op
        >> create_custom_model_version_op
        >> test_dataset_uploading_op
        >> custom_model_test_op
        >> custom_model_test_overall_status_op
        >> if_tests_succeed
        >> [deploy_custom_model_op, custom_model_test_fail]
    )


create_custom_model_pipeline_dag = create_custom_model_pipeline()

if __name__ == "__main__":
    create_custom_model_pipeline_dag.test()



# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# This is proprietary source code of DataRobot, Inc. and its affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections.abc import Sequence
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import datarobot as dr
from airflow.exceptions import AirflowFailException
from airflow.utils.context import Context
from datarobot import Model
from datarobot.insights import ShapImpact
from datarobot.insights import ShapPreview
from datarobot.models import StatusCheckJob

from datarobot_provider.operators.base_datarobot_operator import BaseDatarobotOperator


class GetFeaturesUsedOperator(BaseDatarobotOperator):
    """
    Get the features used for training the model.

    Args:
        project_id (str): DataRobot project ID.
        model_id (str): DataRobot model ID.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.

    Returns:
        List[str]: The names of the features used in the model.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "model_id",
    ]

    def __init__(
        self,
        *,
        project_id: str,
        model_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.model_id = model_id

    def validate(self) -> None:
        if not self.project_id:
            raise ValueError("project_id is required.")

        if not self.model_id:
            raise ValueError("model_id is required.")

    def execute(self, context: Context) -> List[str]:
        model = dr.models.Model.get(self.project_id, self.model_id)
        features = model.get_features_used()

        return features


class ComputeFeatureImpactOperator(BaseDatarobotOperator):
    """
    Creates Feature Impact job in DataRobot.

    Args:
        project_id (str): DataRobot project ID.
        model_id (str): DataRobot model ID.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.

    Returns:
        str: Feature Impact job ID.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "model_id",
    ]

    def __init__(
        self,
        *,
        project_id: str,
        model_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.model_id = model_id

    def validate(self) -> None:
        if not self.project_id:
            raise ValueError("project_id is required to compute Feature Impact.")

        if not self.model_id:
            raise ValueError("model_id is required to compute Feature Impact.")

    def execute(self, context: Context) -> str:
        model = dr.models.Model.get(self.project_id, self.model_id)

        job = model.request_feature_impact()

        self.log.info(f"Feature Impact Job submitted, job_id={job.id}")

        return job.id


class ComputeFeatureEffectsOperator(BaseDatarobotOperator):
    """
    Submit request to compute Feature Effects for the model.

    Args:
        project_id (str): DataRobot project ID.
        model_id (str): DataRobot model ID.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.

    Returns:
        str: Feature Effects job ID.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "model_id",
    ]

    def __init__(
        self,
        *,
        project_id: str,
        model_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.model_id = model_id

    def validate(self) -> None:
        if not self.project_id:
            raise ValueError("project_id is required to compute Feature Effects.")

        if not self.model_id:
            raise ValueError("model_id is required to compute Feature Effects.")

    def execute(self, context: Context) -> str:
        model = dr.models.Model.get(self.project_id, self.model_id)

        job = model.request_feature_effect()

        self.log.info(f"Feature Effects Job submitted, job_id={job.id}")

        return job.id


class ComputeShapPreviewOperator(BaseDatarobotOperator):
    """
    Creates SHAP preview job in DataRobot.

    Parameters
    ----------
    model_id : str
        DataRobot model ID
    datarobot_conn_id : str, optional
        Connection ID, defaults to `datarobot_default`

    Returns
    -------
    str
        SHAP preview result id
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "model_id",
    ]

    def __init__(
        self,
        *,
        model_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.model_id = str(model_id)

    def validate(self) -> None:
        if not self.model_id:
            raise AirflowFailException("The `model_id` parameter is required.")

    def execute(self, context: Context) -> str:
        job: StatusCheckJob = ShapPreview.compute(entity_id=self.model_id)
        return job.job_id


class ComputeShapImpactOperator(BaseDatarobotOperator):
    """
    Creates SHAP impact job in DataRobot.

    Parameters
    ----------
    model_id : str
        DataRobot model ID
    datarobot_conn_id : str, optional
        Connection ID, defaults to `datarobot_default`

    Returns
    -------
    str
        SHAP impact job id
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "model_id",
    ]

    def __init__(
        self,
        *,
        model_id: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.model_id = str(model_id)

    def validate(self) -> None:
        if not self.model_id:
            raise AirflowFailException("The `model_id` parameter is required.")

    def execute(self, context: Context) -> str:
        job: StatusCheckJob = ShapImpact.compute(entity_id=self.model_id)
        return job.job_id


class GetRocCurveInsightOperator(BaseDatarobotOperator):
    """
    Creates ROC curve insight data.

    Args:
        project_id (str): DataRobot model ID.
        model_id (str): DataRobot model ID.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.

    Returns:
        Dict[str, Any]: Roc curve points, positive class predictions, negative class predictions.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "model_id",
    ]

    def __init__(
        self,
        *,
        project_id: str,
        model_id: str,
        source: Optional[str] = "validation",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = str(project_id)
        self.model_id = str(model_id)
        self.source = str(source)

    def validate(self) -> None:
        if not self.project_id:
            raise AirflowFailException("The `project_id` parameter is required.")
        if not self.model_id:
            raise AirflowFailException("The `model_id` parameter is required.")

    def get_model(self) -> Model:
        model: Model = Model.get(self.project_id, self.model_id)
        if not model:
            raise AirflowFailException(
                f"Model with id {self.model_id} not found in project {self.project_id}."
            )
        return model

    def execute(self, context: Context) -> Dict[str, Any]:
        model = self.get_model()
        roc_data = model.get_roc_curve(source=self.source)
        return {
            "roc_points": roc_data.roc_points,
            "positive_class_predictions": roc_data.positive_class_predictions,
            "negative_class_predictions": roc_data.negative_class_predictions,
        }


class GetLiftChartInsightOperator(BaseDatarobotOperator):
    """
    Creates Lift Chart insight data.

    Args:
        project_id (str): DataRobot model ID.
        model_id (str): DataRobot model ID.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.

    Returns:
        List[Dict[str, Any]]: List of the lift chart bins containing the actuals and predicted values.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "model_id",
    ]

    def __init__(
        self,
        *,
        project_id: str,
        model_id: str,
        source: Optional[str] = "validation",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = str(project_id)
        self.model_id = str(model_id)
        self.source = str(source)

    def validate(self) -> None:
        if not self.project_id:
            raise AirflowFailException("The `project_id` parameter is required.")
        if not self.model_id:
            raise AirflowFailException("The `model_id` parameter is required.")

    def get_model(self) -> Model:
        model: Model = Model.get(self.project_id, self.model_id)
        if not model:
            raise AirflowFailException(
                f"Model with id {self.model_id} not found in project {self.project_id}."
            )
        return model

    def execute(self, context: Context) -> List[Dict[str, Any]]:
        model = self.get_model()
        lift_chart = model.get_lift_chart(source=self.source)
        return lift_chart.bins


class GetResidualsChartInsightOperator(BaseDatarobotOperator):
    """
    Creates Residuals Chart insight data.

    Args:
        project_id (str): DataRobot model ID.
        model_id (str): DataRobot model ID.
        datarobot_conn_id (str, optional): Connection ID, defaults to `datarobot_default`.

    Returns:
        Dict[str, float]: The residual chart values for the model.
    """

    # Specify the arguments that are allowed to parse with jinja templating
    template_fields: Sequence[str] = [
        "project_id",
        "model_id",
    ]

    def __init__(
        self,
        *,
        project_id: str,
        model_id: str,
        source: Optional[str] = "validation",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = str(project_id)
        self.model_id = str(model_id)
        self.source = str(source)

    def validate(self) -> None:
        if not self.project_id:
            raise AirflowFailException("The `project_id` parameter is required.")
        if not self.model_id:
            raise AirflowFailException("The `model_id` parameter is required.")

    def get_model(self) -> Model:
        model: Model = Model.get(self.project_id, self.model_id)
        if not model:
            raise AirflowFailException(
                f"Model with id {self.model_id} not found in project {self.project_id}."
            )
        return model

    def execute(self, context: Context) -> Dict[str, float]:
        model = self.get_model()
        residuals = model.get_residuals_chart(source=self.source)
        return {
            "residual_mean": residuals.residual_mean,
            "coefficient_of_determination": residuals.coefficient_of_determination,
            "standard_deviation": residuals.standard_deviation,
        }
